#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
from sirius_data._constants import c, arcmin_to_rad
import scipy.constants
from scipy.special import j1, jn
from numba import jit
import numba_scipy.special
import numba

# The twiddle factor is needed to improve agreement between CASA PBMATH airy disk and SiRIUS. 
# This arrises because CASA makes use of truncated constants. The twiddle factor 0.9998277835716939 is very close to 1. 
# Even with the twiddle there will not be machine precission agreement, because CASA calculates a 1D PB at 1 GHz for 10000 points and then frequency scales and interpolates to desired value.
# CASA also discards the fractional part when calculating the index (use Int instead of (int(np.floor(r/r_inc + 0.5)).
# See PBMath1D.cc Complex version ImageInterface<Complex>&PBMath1D::apply
# SkyComponent&PBMath1D::apply makes a small angle approximation while ImageInterface<Complex>&PBMath1D::apply does not. We have decided to not make the small angle approximation.
casa_twiddle = (180*7.016*c)/((np.pi**2)*(10**9)*1.566*24.5) # 0.9998277835716939

@jit(nopython=True,nogil=True)
def _casa_airy_beam(l,m,freq_chan,dish_diameter, blockage_diameter, ipower, max_rad_1GHz, n_sample=10000):
    """
    Airy disk function to model the antenna response, that has been adapted to match the PBMATH class in CASA. CASA PBMath does not directly calculate the airy function. It generates 10000 points at 1 GHz and then scales and samples from that when apply_pb/vp is called. As part of the scaling someone truncated some factors, so to get a python function that matches CASA a twiddle factor: casa_twiddle = (1807.016c)/((np.pi**2)(10**9)*1.56624.5)=0.9998277835716939 is added. It also does integer rounding. This function should agree with CASA up to 7 significant figures.
    
    Parameters
    ----------
    l: float, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    m: float, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    freq_chan: float, Hz
        Frequency.
    dish_diameter: float, meters
        The diameter of the dish.
    blockage_diameter: float, meters
        The central blockage of the dish.
    ipower: int
        ipower = 1 single dish response.
        ipower = 2 baseline response for identical dishes.
    max_rad_1GHz: float, radians
        The max radius from which to sample scaled to 1 GHz.
        This value can be found in sirius_data.dish_models_1d.airy_disk.
        For example the Alma dish model (sirius_data.dish_models_1d.airy_disk import alma)
        is alma = {'func': 'airy', 'dish_diam': 10.7, 'blockage_diam': 0.75, 'max_rad_1GHz': 0.03113667385557884}.
    n_sample=10000
        The sampling used in CASA for PB math.
    Returns
    -------
    val : float
        The dish response.
    """
    r_max = max_rad_1GHz/(freq_chan/10**9)
    k = (2*np.pi*freq_chan)/c
    aperture = dish_diameter/2

    if n_sample is not None:
        r = np.sqrt(l**2 + m**2)
        r_inc = ((r_max)/(n_sample-1))
        r = (int(r/r_inc)*r_inc)*aperture*k #Int rounding instead of r = (int(np.floor(r/r_inc + 0.5))*r_inc)*aperture*k
        r = r*casa_twiddle
    else:
        r = np.arcsin(np.sqrt(l**2 + m**2)*k*aperture)
        
    if (r != 0):
        if blockage_diameter==0.0:
            return (2.0*j1(r)/r)**ipower
        else:
            area_ratio = (dish_diameter/blockage_diameter)**2
            length_ratio = (dish_diameter/blockage_diameter)
            return ((area_ratio * 2.0 * j1(r)/r   - 2.0 * j1(r * length_ratio)/(r * length_ratio) )/(area_ratio - 1.0))**ipower
    else:
        return 1


#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _3d_casa_airy_beam(l,m,freq_chan,dish_diameter, blockage_diameter, ipower, max_rad_1GHz, n_sample=10000):
    """
    3-dimensional Airy disk function to model the antenna response, that has been adapted to match the PBMATH class in CASA. CASA PBMath does not directly calculate the airy function. It generates 10000 points at 1 GHz and then scales and samples from that when apply_pb/vp is called. As part of the scaling someone truncated some factors, so to get a python function that matches CASA a twiddle factor: casa_twiddle = (1807.016c)/((np.pi**2)(10**9)*1.56624.5)=0.9998277835716939 is added. It also does integer rounding. This function should agree with CASA up to 7 significant figures.
    Parameters
    ----------
    l: float np.array, n_pix_x, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    m: float np.array, n_pix_y, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    freq_chan: float np.float, n_chan, Hz
        Frequency.
    dish_diameter: float, meters
        The diameter of the dish.
    blockage_diameter: float, meters
        The central blockage of the dish.
    ipower: int
        ipower = 1 single dish response.
        ipower = 2 baseline response for identical dishes.
    max_rad_1GHz: float, radians
        The max radius from which to sample scaled to 1 GHz.
        This value can be found in sirius_data.dish_models_1d.airy_disk.
        For example the Alma dish model (sirius_data.dish_models_1d.airy_disk import alma)
        is alma = {'func': 'airy', 'dish_diam': 10.7, 'blockage_diam': 0.75, 'max_rad_1GHz': 0.03113667385557884}.
    n_sample=10000
        The sampling used in CASA for PB math.
    Returns
    -------
    beam : float np.array, [n_chan,n_pix_x,n_pix_y]
        The dish response.
    """
    beam = np.zeros(freq_chan.shape + l.shape + m.shape,dtype=numba.float64)
    
    for i_freq,freq in enumerate(freq_chan):
            r_max = max_rad_1GHz/(freq/10**9)
            k = (2*np.pi*freq)/c
            aperture = dish_diameter/2

            for i_l, l_val in enumerate(l):
                for i_m, m_val in enumerate(m):
                    if n_sample is not None:
                        r = np.sqrt(l_val**2 + m_val**2)
                        r_inc = ((r_max)/(n_sample-1))
                        r = (int(r/r_inc)*r_inc)*aperture*k #Int rounding instead of r = (int(np.floor(r/r_inc + 0.5))*r_inc)*aperture*k
                        r = r*casa_twiddle
                    else:
                        r = np.sqrt(l_val**2 + m_val**2)*k*aperture
                        
                    if (r != 0):
                        if blockage_diameter==0.0:
                            beam[i_freq,i_l,i_m] = (2.0*j1(r)/r)**ipower
                        else:
                            area_ratio = (dish_diameter/blockage_diameter)**2
                            length_ratio = (dish_diameter/blockage_diameter)
                            beam[i_freq,i_l,i_m] = ((area_ratio * 2.0 * j1(r)/r   - 2.0 * j1(r * length_ratio)/(r * length_ratio) )/(area_ratio - 1.0))**ipower
                    else:
                        beam[i_freq,i_l,i_m] = 1
    return beam
    
 

#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _airy_beam(l,m,freq_chan,dish_diameter, blockage_diameter, ipower):
    """
    Airy disk function to model the antenna response.
    
    Parameters
    ----------
    l: float, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    m: float, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    freq_chan: float, Hz
        Frequency.
    dish_diameter: float, meters
        The diameter of the dish.
    blockage_diameter: float, meters
        The central blockage of the dish.
    ipower: int
        ipower = 1 single dish response.
        ipower = 2 baseline response for identical dishes.
    Returns
    -------
    val : float
        The dish response.
    """

    k = (2*np.pi*freq_chan)/c
    aperture = dish_diameter/2
    r = np.sqrt(l**2 + m**2)*k*aperture
        
    if (r != 0):
        if blockage_diameter==0.0:
            return (2.0*j1(r)/r)**ipower
        else:
            e = blockage_diameter/dish_diameter
            return (( 2.0 * j1(r)/r   - 2.0 * e * j1(r * e)/r )/(1.0 - e**2))**ipower
            #Changed r_grid to r ^^^
    else:
        return 1
 
#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _3d_airy_beam(l,m,freq_chan,dish_diameter, blockage_diameter, ipower, max_rad_1GHz):
    """
    3-dimensional Airy disk function to model the antenna response.
    Parameters
    ----------
    l: float np.array, n_pix_x, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    m: float np.array, n_pix_y, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    freq_chan: float np.float, n_chan, Hz
        Frequency.
    dish_diameter: float, meters
        The diameter of the dish.
    blockage_diameter: float, meters
        The central blockage of the dish.
    ipower: int
        ipower = 1 single dish response.
        ipower = 2 baseline response for identical dishes.
    max_rad_1GHz: float, radians
        The max radius from which to sample scaled to 1 GHz.
        This value can be found in sirius_data.dish_models_1d.airy_disk.
        For example the Alma dish model (sirius_data.dish_models_1d.airy_disk import alma)
        is alma = {'func': 'airy', 'dish_diam': 10.7, 'blockage_diam': 0.75, 'max_rad_1GHz': 0.03113667385557884}.
    Returns
    -------
    beam : float np.array, [n_chan,n_pix_x,n_pix_y]
        The dish response.
    """
    beam = np.zeros(freq_chan.shape + l.shape + m.shape,dtype=numba.float64)
    
    for i_freq,freq in enumerate(freq_chan):
            r_max = max_rad_1GHz/(freq/10**9)
            k = (2*np.pi*freq)/c
            aperture = dish_diameter/2

            for i_l, l_val in enumerate(l):
                for i_m, m_val in enumerate(m):
           
                    r = np.sqrt(l_val**2 + m_val**2)*k*aperture
                        
                    if (r != 0):
                        if blockage_diameter==0.0:
                            beam[i_freq,i_l,i_m] = (2.0*j1(r)/r)**ipower
                        else:
                            e = blockage_diameter/dish_diameter
                            beam[i_freq,i_l,i_m] = (( 2.0 * j1(r)/r   - 2.0 * e * j1(r * e)/r )/(1.0 - e**2))**ipower
                    else:
                        beam[i_freq,i_l,i_m] = 1
    return beam
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
 #######################################
@jit(nopython=True,nogil=True)
def _poly_beam(l,m,freq,coef, ipower, max_rad_1GHz, n_sample=10000):
    """
    3-dimensional
    
    Parameters
    ----------
    l: float np.array, n_pix_x, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    m: float np.array, n_pix_y, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    coef: float,
        
    ipower: int
        ipower = 1 single dish response.
        ipower = 2 baseline response for identical dishes.
    Returns
    -------
    beam : float np.array, [n_chan,n_pix_x,n_pix_y]
        The dish response.
    """
    r = np.sqrt(l**2 + m**2)
    beam = 0.0
    r_max = max_rad_1GHz/(freq/10**9)
    
    if n_sample is not None:
        r_inc = ((r_max)/(n_sample-1))
        r = (int(r/r_inc)*r_inc) #Int rounding
        
    r = r*(freq/10**9)/arcmin_to_rad
    for i_coef,cf in enumerate(coef[0,:]):
        beam = beam + cf*(r)**(2*i_coef)
    beam = beam**(0.5*ipower)
    return beam
    

@jit(nopython=True,nogil=True)
def _3d_poly_beam(l,m,freq_chan,coef, ipower, max_rad_1GHz):
    """
    3-dimensional
    
    Parameters
    ----------
    l: float np.array, n_pix_x, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    m: float np.array, n_pix_y, radians
        Coordinate of a point on the image plane (the synthesis projected ascension and declination).
    coef: float,
        
    ipower: int
        ipower = 1 single dish response.
        ipower = 2 baseline response for identical dishes.
    Returns
    -------
    beam : float np.array, [n_chan,n_pix_x,n_pix_y]
        The dish response.
    """
    n_chan = coef.shape[0]
    beam = np.zeros((n_chan,) + l.shape + m.shape,dtype=numba.float64)

    for i_c in range(n_chan):
            f = freq_chan[i_c]/10**9
            r_max = max_rad_1GHz/f
            for i_l, l_val in enumerate(l):
                for i_m, m_val in enumerate(m):
                    r = np.sqrt(l_val**2 + m_val**2)
                    
                    if r < r_max:
                        r = r*f/arcmin_to_rad
                        for i_coef,cf in enumerate(coef[i_c,0,:]):
                            beam[i_c,i_l,i_m] = beam[i_c,i_l,i_m] + cf*(r)**(2*i_coef)
                        beam[i_c,i_l,i_m] = beam[i_c,i_l,i_m]**(0.5*ipower)
    return beam


    
    
    
    
    
    
    
    
    
    
    
    
    
##############################################################################################################
#Non-jitted version:
def _casa_airy_beam_njit(lmn,freq_chan,beam_parms):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) or (lmn[1] != 0):
        dish_diameter = beam_parms['dish_diameter']
        blockage_diameter = beam_parms['blockage_diameter']
        ipower = beam_parms['ipower']
        
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        
        if blockage_diameter==0.0:
            return (2.0*jn(1,r)/r)**ipower
        else:
            area_ratio = (dish_diameter/blockage_diameter)**2
            length_ratio = (dish_diameter/blockage_diameter)
            return ((area_ratio * 2.0 * jn(1,r)/r   - 2.0 * jn(1, r * length_ratio)/(r * length_ratio) )/(area_ratio - 1.0))**ipower
    else:
        return 1


def _airy_beam_njit(lmn,freq_chan,beam_parms):
    #print('lmn is',lmn)
    
    if (lmn[0] != 0) or (lmn[1] != 0):
        dish_diameter = beam_parms['dish_diameter']
        blockage_diameter = beam_parms['blockage_diameter']
        ipower = beam_parms['ipower']
        
        k = (2*np.pi*freq_chan)/c
        
        aperture = dish_diameter/2
        r = np.sqrt(lmn[0]**2 + lmn[1]**2)*k*aperture
        
        if blockage_diameter==0.0:
            return (2.0*jn(1,r)/r)**ipower
        else:
            e = blockage_diameter/dish_diameter
            return (( 2.0 * jn(1,r)/r   - 2.0 * e * jn(1, r * e)/r )/(1.0 - e**2))**ipower
    else:
        return 1



#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Four different implementations for calculating the parallactic angle have been included. Each version produces a slightly different answer.
# The reason for the differences has not been verified, however, we suspect they have to do with different earth models and earth movement corrections that are applied by the different software packages.
# Here is an interesting discussion about different packages giving differing parallactic angles: https://github.com/ratt-ru/codex-africanus/issues/21


import numpy as np
from numba import jit
import numba

#AZEL vs AZELGEO
#Dir FK5 (J2000) vs ICRS


def _calc_parallactic_angles_astropy(times, observing_location, direction, dir_frame='FK5', zenith_frame='FK5'):
    """
    Converts a direction and zenith (frame FK5) to a topocentric Altitude-Azimuth (https://docs.astropy.org/en/stable/api/astropy.coordinates.AltAz.html) frame centered at the observing_location (frame ITRF) for a UTC time. The parallactic angles is calculated as the position angle of the Altitude-Azimuth direction and zenith.
    
    Parameters
    ----------
    times: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        UTC time series. Example '2019-10-03T19:00:00.000'.
    observing_location: int np.array, [3], [x,y,z], meters
        ITRF geocentric coordinates.
    direction: float np.array, [n_time,2], [time,ra,dec], radians
        The pointing direction.
    Returns
    -------
    parallactic_angles: float np.array, [n_time], radians
        An array of parallactic angles.
    """
    import astropy.units as u
    import astropy.coordinates as coord
    
    observing_location = coord.EarthLocation.from_geocentric(x=observing_location[0]*u.m, y=observing_location[1]*u.m, z=observing_location[2]*u.m)
    
    direction = coord.SkyCoord(ra=direction[:,0]*u.rad, dec=direction[:,1]*u.rad, frame=dir_frame.lower())
    zenith = coord.SkyCoord(0, 90, unit=u.deg, frame=zenith_frame.lower())
    
    altaz_frame = coord.AltAz(location=observing_location, obstime=times)
    zenith_altaz = zenith.transform_to(altaz_frame)
    direction_altaz = direction.transform_to(altaz_frame)
    
    return direction_altaz.position_angle(zenith_altaz).value
    
    '''
    cirs_frame = coord.CIRS(obstime=times)
    zenith_cirs = zenith.transform_to(cirs_frame)
    direction_cirs = direction.transform_to(cirs_frame)

    altaz_frame = coord.AltAz(location=observing_location, obstime=times)
    zenith_altaz = zenith_cirs.transform_to(altaz_frame)
    direction_altaz = direction_cirs.transform_to(altaz_frame)
    
    return direction_altaz.position_angle(zenith_altaz).value
    '''
    
def _calc_parallactic_angles_astropy2(times, observing_location, direction, dir_frame='FK5'):
    """
    Calculates a set of parallactic angles for a array of time strings for a given earth obeserving location and direction.
    Based on the Astropy implementation in https://github.com/ARDG-NRAO/plumber/blob/master/plumber/sky.py.

    Parameters
    ----------
    times: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        UTC time series. Example '2019-10-03T19:00:00.000'.
    observing_location: int np.array, [3], [x,y,z], meters
        ITRF geocentric coordinates.
    direction: float np.array, [n_time,2], [time,ra,dec], radians
        The pointing direction.
    Returns
    -------
    parallactic_angles: float np.array, [n_time], radians
        An array of parallactic angles.
    """
    from astroplan import Observer
    import astropy.units as u
    import astropy.coordinates as coord
    from astropy.time import Time
    
    from casatools import measures
    me = measures()
    
    observing_location = coord.EarthLocation.from_geocentric(x=observing_location[0]*u.m, y=observing_location[1]*u.m, z=observing_location[2]*u.m)
    
    direction = coord.SkyCoord(ra=direction[:,0]*u.rad, dec=direction[:,1]*u.rad, frame=dir_frame.lower())
    
    observer = Observer(location=observing_location, name='tel', timezone='UTC')
    
    times = Time(times)

    return observer.parallactic_angle(times, direction).value
    

def _calc_parallactic_angles_casa(times, observing_location, direction, frame='AZEL', dir_frame='FK5', zenith_frame='HADEC'):
    """
    Calculates a set of parallactic angles for a array of time strings for a given earth obeserving location and direction.
    Makes use of casacore measures posangle function (https://casacore.github.io/python-casacore/casacore_measures.html#casacore.measures.measures.posangle).

    Parameters
    ----------
    times: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        UTC time series. Example '2019-10-03T19:00:00.000'.
    observing_location: int np.array, [3], [x,y,z], meters
        ITRF geocentric coordinates.
    direction: float np.array, [n_time,2], [time,ra,dec], radians
        The pointing direction.
    Returns
    -------
    parallactic_angles: float np.array, [n_time], radians
        An array of parallactic angles.
    """
    from casacore.measures import measures
    from casacore.quanta import quantity as qq
    me = measures()
    
    n_time = len(times)
    f_pc_time = n_time if direction.shape[0] == 1 else 1
    
    observing_location = me.position('ITRF',
                    qq(observing_location[0],'m'),
                    qq(observing_location[1],'m'),
                    qq(observing_location[2],'m'))
    
    if dir_frame=='FK5':
        dir_frame = 'J2000' #Change to CASA convention.
        
    if zenith_frame=='FK5':
        zenith_frame = 'J2000'
        
    if frame == 'FK5':
        frame='J2000'
    
    parallactic_angles = np.zeros(n_time)
    for i,t in enumerate(times):
        me = measures()
        me.doframe(observing_location)
        me.doframe(me.epoch('utc', t))
        
        zenith = me.measure(me.direction(rf=zenith_frame, v0='0deg', v1='90deg'),frame)
        pointing_dir = me.measure(me.direction(dir_frame,qq(direction[i//f_pc_time,0], 'rad'),qq(direction[i//f_pc_time,1], 'rad')),frame)
        
        #print(pointing_dir)
 
        parallactic_angles[i] = me.posangle(pointing_dir, zenith).get_value("rad")
 
    return parallactic_angles
    
def _calc_parallactic_angles_casa2(times, observing_location, direction, dir_frame='FK5'):
    """
    Calculates a set of parallactic angles for a array of time strings for a given earth obeserving location and direction.
    Based on the Casa implementation in https://github.com/ARDG-NRAO/plumber/blob/master/plumber/sky.py.

    Parameters
    ----------
    times: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        UTC time series. Example '2019-10-03T19:00:00.000'.
    observing_location: int np.array, [3], [x,y,z], meters
        ITRF geocentric coordinates.
    direction: float np.array, [n_time,2], [time,ra,dec], radians
        The pointing direction.
    Returns
    -------
    parallactic_angles: float np.array, [n_time], radians
        An array of parallactic angles.
    """
    from casacore.measures import measures
    from casacore.quanta import quantity as qq
    me = measures()
    
    n_time = len(times)
    f_pc_time = n_time if direction.shape[0] == 1 else 1
    
    lat = np.arcsin(observing_location[2]/np.sqrt(np.sum(observing_location**2)))
    observing_location = me.position('ITRF',
                    qq(observing_location[0],'m'),
                    qq(observing_location[1],'m'),
                    qq(observing_location[2],'m'))
                    
    parallactic_angles = np.zeros(n_time)
    
    if dir_frame=='FK5':
        dir_frame = 'J2000' #Change to CASA convention.
    
    for i,t in enumerate(times):
        me = measures()
        me.doframe(observing_location)
        tm = me.epoch('utc', t)
        last = me.measure(tm, 'LAST')['m0']['value']
        pointing_dir = me.direction(dir_frame,qq(direction[i//f_pc_time,0], 'rad'),qq(direction[i//f_pc_time,1], 'rad'))
        
        ha = (last - np.floor(last))*24 - (pointing_dir['m0']['value']*12/np.pi) # hours
        ha = ha*np.pi/12.  # rad
        dec = pointing_dir['m1']['value']
        parallactic_angles[i] = np.arctan2(np.cos(lat)*np.sin(ha),(np.sin(lat) * np.cos(dec) - np.cos(lat)*np.sin(dec)*np.cos(ha)))
        
    return parallactic_angles

    
    
def _calc_parallactic_angles_casa3(times, observing_location, direction, frame='AZEL', dir_frame='FK5', zenith_frame='HADEC'):
    """
    Calculates a set of parallactic angles for a array of time strings for a given earth obeserving location and direction.
    Based on: casacore/ms/MSOper/MSDerivedValues.cc::parAngle(),casacore/casa/Quanta/MVDirection.cc::positionAngle.
    This casacore code is used by awproject in CASA tclean. Note that the casacore/measures/Measures/ParAngleMachine.cc is an independant implementation (see _calc_parallactic_angles_casa).

    Parameters
    ----------
    times: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        UTC time series. Example '2019-10-03T19:00:00.000'.
    observing_location: int np.array, [3], [x,y,z], meters
        ITRF geocentric coordinates.
    direction: float np.array, [n_time,2], [time,ra,dec], radians
        The pointing direction.
    Returns
    -------
    parallactic_angles: float np.array, [n_time], radians
        An array of parallactic angles.
    """
    from casacore.measures import measures
    from casacore.quanta import quantity as qq
    from sirius._sirius_utils._coord_transforms import _directional_cosine
    me = measures()
    
    n_time = len(times)
    f_pc_time = n_time if direction.shape[0] == 1 else 1
    
    lat = np.arcsin(observing_location[2]/np.sqrt(np.sum(observing_location**2)))
    observing_location = me.position('ITRF',
                    qq(observing_location[0],'m'),
                    qq(observing_location[1],'m'),
                    qq(observing_location[2],'m'))
    me.doframe(observing_location)
    
    parallactic_angles = np.zeros(n_time)
    
    if dir_frame=='FK5':
        dir_frame = 'J2000' #Change to CASA convention.
        
    if zenith_frame=='FK5':
        zenith_frame = 'J2000'
        
    if frame == 'FK5':
        frame='J2000'
    
    
    for i,t in enumerate(times):
        me = measures()
        me.doframe(observing_location)
        reference_time = me.epoch('utc', t)
        me.doframe(reference_time)
        
        #Create Pole direction
        zenith = me.measure(me.direction(rf=zenith_frame, v0='0deg', v1='90deg'),frame)
        zenith_cosine = _directional_cosine(np.array([[zenith['m0']['value'],zenith['m1']['value']]]))[0,:]

        pointing_dir = me.measure(me.direction(dir_frame,qq(direction[i//f_pc_time,0], 'rad'),qq(direction[i//f_pc_time,1], 'rad')), frame)
        pointing_cosine=_directional_cosine(np.array([[pointing_dir['m0']['value'],pointing_dir['m1']['value']]]))[0,:]
        
        zenith_long = np.arctan2(zenith_cosine[1],zenith_cosine[0])
        pointing_long = np.arctan2(pointing_cosine[1],pointing_cosine[0])
        
        pc_slat1 = pointing_cosine[2];
        zenith_slat2 = zenith_cosine[2];
        zenith_clat2 = np.sqrt(np.abs(1 - zenith_slat2**2))
        
        s1 = (-zenith_clat2 * np.sin(pointing_long-zenith_long))
        c1 = (np.sqrt(np.abs(1.0 - pc_slat1**2))*zenith_slat2 - pc_slat1*zenith_clat2*np.cos(pointing_long-zenith_long));
        parallactic_angles[i] = np.arctan2(s1,c1)
    return parallactic_angles
    

@jit(nopython=True,cache=True,nogil=True)
def _find_optimal_set_angle(ang_vals,val_step,calc_dif=False):
    """
    For an array of angles (ang_vals) find an "optimal" (ie good) subset of angles (subset_ang_vals) so that these angles all lie within val_step of all the values in val_step.
    
    Parameters
    ----------
    ang_vals: float np.array, [n_vals], radians
    val_step: float, radians
    calc_dif: bool
    Returns
    -------
    subset_ang_vals: float np.array, [n_subset_vals], radians
    vals_dif: float np.array, [n_vals], radians
        Optional output if calc_dif is True. The minimum angular distance for each ang_vals with subset_ang_vals.
    vals_indx: int np.array, [n_vals], radians
        Optional output if calc_dif is True. The index for each ang_vals into subset_ang_vals so that the angular distance is a minimum.
    """
    n_vals = len(ang_vals)
    neighbours = np.identity(n_vals,numba.b1)
    #neighbours = np.identity(n_vals,int)

    # Find all the neighbours. A boolean n_vals x n_vals array, called neighbours, is created.
    # If two angles are within val_step they are considered neighbours.
    for ii in range(0,n_vals-1):
        for jj in range(ii+1,n_vals):
            #https://stackoverflow.com/questions/1878907/the-smallest-difference-between-2-angles
            ang_dif = ang_vals[ii]-ang_vals[jj]
            ang_dif = np.abs((ang_dif + np.pi)%(2*np.pi) - np.pi)
            
            if ang_dif <= val_step:
                neighbours[ii,jj] = True
                neighbours[jj,ii] = True
                
    # Create neighbour groups by iteratively selecting the val with the highest neighbours_rank and setting that val and all its neigbours to False (0) in neighbours. Currently, the value of the neighbour with the highest neighbours_rank is used as the center value for the group (other options are commented out: median and mean). When all values are False (0) in neighbours the process ends.
    subset_ang_vals = [42.0] #Dummy value to let numba know what dtype of list is.
    lonely_neighbour = True
    while lonely_neighbour:
        #if True:
        neighbours_rank = np.sum(neighbours,axis=1) #Each val has a neighbours_rank equal to all the vals that are within val_step.
        highest_ranked_neighbour_indx = np.argmax(neighbours_rank)
        
        if neighbours_rank[highest_ranked_neighbour_indx]==0:
            lonely_neighbour = False
        else:
            group_members = np.where(neighbours[highest_ranked_neighbour_indx,:]==1)[0]
            subset_ang_vals.append(ang_vals[highest_ranked_neighbour_indx])
            #subset_ang_vals.append(np.median(ang_vals[neighbours[highest_ranked_neighbour_indx,:]])) #best stats
            #subset_ang_vals.append(np.mean(ang_vals[neighbours[highest_ranked_neighbour_indx,:]])) #?
            
            for group_member in group_members:
                for ii in range(n_vals):
                    neighbours[group_member,ii] = 0
                    neighbours[ii,group_member] = 0
                    
    subset_ang_vals.pop(0)
    subset_ang_vals = np.array(subset_ang_vals)
    
    if calc_dif:
        vals_dif = np.zeros(n_vals,numba.f8)
        vals_indx = np.zeros(n_vals,numba.f8)
        
        # Calculate the min angular distance between subset_ang_vals and ang_vals for each value in ang_vals.
        # The index of the value in subset_ang_vals that gives the min angular distance is also stored in vals_indx.
        for ii in range(n_vals):
            min_dif = 42.0 #Dummy value to let numba know the dtype.
            indx = -42 #Dummy value to let numba know the dtype.
            for jj in range(len(subset_ang_vals)):
                #https://stackoverflow.com/questions/1878907/the-smallest-difference-between-2-angles
                ang_dif = ang_vals[ii]-subset_ang_vals[jj]
                ang_dif = np.abs((ang_dif + np.pi)%(2*np.pi) - np.pi)
                
                if min_dif > ang_dif:
                    min_dif = ang_dif
                    indx = jj
            
            vals_dif[ii] = min_dif
            vals_indx[ii] = indx
        return subset_ang_vals, vals_dif, vals_indx
    else:
        return subset_ang_vals, np.array([0.0]), np.array([0.0])






 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import numpy as np
import xarray as xr
import copy
from sirius._sirius_utils._beam_utils import _calc_ant_jones, _calc_resolution, _pol_code_to_index, _index_to_pol_code
from sirius._sirius_utils._calc_parallactic_angles import _calc_parallactic_angles_astropy, _find_optimal_set_angle
from sirius._parm_utils._check_beam_parms import _check_beam_parms
from sirius._sirius_utils._beam_funcs import _3d_casa_airy_beam, _3d_airy_beam, _3d_poly_beam
from sirius_data._constants import map_mueler_to_pol, c

def calc_zpc_beam(zpc_xds,parallactic_angles,freq_chan,beam_parms,check_parms=True):
    """
    Calculates an antenna apertures from Zernike polynomial coefficients, and then Fourier transform it to obtain the antenna beam image.
    The beam image dimensionality is [pa (paralactic angle), chan (channel), pol (polarization), l (orthographic/synthesis projection of directional cosine), m (orthographic/synthesis projection of directional cosine)].

    Parameters
    ----------
    zpc_xds: xr.Dataset
        A Zernike polynomial coefficient xr.Datasets. Available models can be found in sirius_data/zernike_dish_models/data.
    parallactic_angles: float np.array, [n_pa], radians
        An array of the parallactic angles for which to calculate the antenna beams.
    freq_chan: float np.array, [n_chan], Hz
        Channel frequencies.
    beam_parms: dict
    beam_parms['mueller_selection']: int np.array, default=np.array([ 0, 5, 10, 15])
        The elements in the 4x4 beam Mueller matrix to use. The elements are numbered row wise.
        For example [ 0, 5, 10, 15] are the diagonal elements.
    beam_parms['pa_radius']: float, default=0.2, radians
        The change in parallactic angle that will trigger the calculation of a new beam when using Zernike polynomial aperture models.
    beam_parms['image_size']: int np.array, default=np.array([1000,1000])
        This parameter should rarely be modified. Size of the beam image generated from the Zernike polynomial coefficients.
    beam_parms['fov_scaling']: int, default=1.2
        This parameter should rarely be modified. Used to determine the cell size of the beam image so that it lies within the image that is generated.
    beam_parms['zernike_freq_interp']: str, default='nearest', options=['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic']
        What interpolation method to use for Zernike polynomial coefficients.
    check_parms: bool
        Check input parameters and asign defaults.
        
    Returns
    -------
    J_xds: xr.Dataset
        An xds that contains the image of the per antenna beam as a function of [pa (paralactic angle), chan (channel), pol (polarization), l (orthographic/synthesis projection of directional cosine), m (orthographic/synthesis projection of directional cosine)]. Should not be confused with the primary beam, which is the beam for a baseline and is equal to the product of two antenna beams.
    """
    _beam_parms = copy.deepcopy(beam_parms)
    
    if check_parms: assert(_check_beam_parms(_beam_parms)), "######### ERROR: beam_parms checking failed."
    
    #min_delta = _calc_resolution(freq_chan,zpc_xds.dish_diam,_beam_parms)
    #_beam_parms['cell_size'] = np.array([-min_delta,min_delta]) #- sign?
    delta = (zpc_xds.attrs['max_rad_1GHz']/np.min(freq_chan/10**9))/_beam_parms['image_size']
    _beam_parms['cell_size'] = np.array([-delta[0],delta[1]])*_beam_parms['fov_scaling'] #- sign?
    
    map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])
    _beam_parms['needed_pol'] = np.unique(np.ravel(map_mueler_to_pol[_beam_parms['mueller_selection']]))
    
    assert (0 in _beam_parms['mueller_selection']) or (15 in _beam_parms['mueller_selection']), "Mueller element 0 or 15 must be selected."
    
    J = _calc_ant_jones(zpc_xds,freq_chan,parallactic_angles,_beam_parms)
    
    image_size = _beam_parms['image_size']
    image_center = image_size//2
    cell_size = _beam_parms['cell_size']
    
    image_center = np.array(image_size)//2
    l = np.arange(-image_center[0], image_size[0]-image_center[0])*cell_size[0]
    m = np.arange(-image_center[1], image_size[1]-image_center[1])*cell_size[1]

    coords = {'chan':freq_chan, 'pa': parallactic_angles, 'pol': _index_to_pol_code(_beam_parms['needed_pol'],zpc_xds.pol.values),'l':l,'m':m}
    
    J_xds = xr.Dataset()
    J_xds = J_xds.assign_coords(coords)
    
    J_xds['J'] = xr.DataArray(J, dims=['pa','chan','pol','l','m'])
    
    return J_xds
    
def calc_beam(func_parms,freq_chan,beam_parms,check_parms=True):

    _beam_parms = copy.deepcopy(beam_parms)
    _func_parms = copy.deepcopy(func_parms)
    if check_parms: assert(_check_beam_parms(_beam_parms)), "######### ERROR: beam_parms checking failed."
    
    
    #min_delta = _calc_resolution(freq_chan,_func_parms['dish_diam'],_beam_parms)
    #_beam_parms['cell_size'] = np.array([-min_delta,min_delta]) #- sign?
    delta = (func_parms['max_rad_1GHz']/np.min(freq_chan/10**9))/_beam_parms['image_size']
    _beam_parms['cell_size'] = np.array([-delta[0],delta[1]])*_beam_parms['fov_scaling'] #- sign?
    print(_beam_parms['image_size'],_beam_parms['cell_size'],_beam_parms['fov_scaling'],func_parms['max_rad_1GHz'])
    
    image_size = _beam_parms['image_size']
    image_center = image_size//2
    cell_size = _beam_parms['cell_size']
    
    image_center = np.array(image_size)//2
    l = np.arange(-image_center[0], image_size[0]-image_center[0])*cell_size[0]
    m = np.arange(-image_center[1], image_size[1]-image_center[1])*cell_size[1]
    
    ipower = 1
    if _func_parms['func'] == 'casa_airy':
        J = _3d_casa_airy_beam(l,m,freq_chan,_func_parms['dish_diam'], _func_parms['blockage_diam'],ipower,_func_parms['max_rad_1GHz'])
    elif _func_parms['func'] == 'airy':
        J = _3d_airy_beam(l,m,freq_chan,_func_parms['dish_diam'], _func_parms['blockage_diam'], ipower)

    coords = {'chan':freq_chan, 'pa': [0], 'pol': [0],'l':l,'m':m}
    
    J_xds = xr.Dataset()
    J_xds = J_xds.assign_coords(coords)
    
    J_xds['J'] = xr.DataArray(J[None,:,None,:,:], dims=['pa','chan','pol','l','m'])
    
    return J_xds
    
def calc_bpc_beam(bpc_xds,freq_chan,beam_parms,check_parms=True):

    _beam_parms = copy.deepcopy(beam_parms)
    if check_parms: assert(_check_beam_parms(_beam_parms)), "######### ERROR: beam_parms checking failed. "
    
    
    #min_delta = _calc_resolution(freq_chan,bpc_xds.attrs['dish_diam'],_beam_parms)
    #_beam_parms['cell_size'] = np.array([-min_delta,min_delta]) #- sign?
    delta   = (bpc_xds.attrs['max_rad_1GHz']/np.min(freq_chan/10**9))/_beam_parms['image_size']
    _beam_parms['cell_size'] = np.array([-delta[0],delta[1]])*_beam_parms['fov_scaling']
    
    #print('cell_size ',_beam_parms['cell_size'])
    
    image_size = _beam_parms['image_size']
    image_center = image_size//2
    cell_size = _beam_parms['cell_size']
    
    image_center = np.array(image_size)//2
    l = np.arange(-image_center[0], image_size[0]-image_center[0])*cell_size[0]
    m = np.arange(-image_center[1], image_size[1]-image_center[1])*cell_size[1]
    
    ipower = 1
    J = _3d_poly_beam(l,m,freq_chan,bpc_xds.BPC.values,ipower,bpc_xds.attrs['max_rad_1GHz'])

    coords = {'chan':freq_chan, 'pa': [0], 'pol': [0],'l':l,'m':m}
    
    J_xds = xr.Dataset()
    J_xds = J_xds.assign_coords(coords)
    
    J_xds['J'] = xr.DataArray(J[None,:,None,:,:], dims=['pa','chan','pol','l','m'])
    
    return J_xds


def evaluate_beam_models(beam_models,time_str,freq_chan,phase_center_ra_dec,site_location,beam_parms,check_parms=True):
    """
    Loops over beam_models and converts each Zernike polynomial coefficient xr.Datasets to an antenna beam image. The beam image dimensionality is [pa (paralactic angle), chan (channel), pol (polarization), l (orthographic/synthesis projection of directional cosine), m (orthographic/synthesis projection of directional cosine)]. The parallactic angles are also calculated for each date-time in time_str at the site_location and with a right ascension declination in phase_center_ra_dec. A subset of parallactic angles are used in the pa coordinate of the beam image, where all pa values are within beam_parms['pa_radius'] radians.

    Parameters
    ----------
    beam_models: list
        List of beam models to use. Beam models can be any combination of function parameter dictionaries, image xr.Datasets or Zernike polynomial coefficient xr.Datasets (models can be found in sirius_data/zernike_dish_models/data).
    time_str: str np.array, [n_time], 'YYYY-MM-DDTHH:MM:SS.SSS'
        Time series. Example '2019-10-03T19:00:00.000'.
    freq_chan: float np.array, [n_chan], Hz
        Channel frequencies.
    phase_center_ra_dec: float np.array, [n_time, 2], (singleton: n_time), radians
        Phase center of array.
    site_location: dict
        A dictionary with the location of telescope. For example [{'m0': {'unit': 'm', 'value': -1601185}, 'm1': {'unit': 'm', 'value': -5041977}, 'm2': {'unit': 'm', 'value': 3554875}, 'refer': 'ITRF', 'type': 'position'}]. The site location of telescopes can be found in site_pos attribute of the xarray dataset of the radio telescope array layout (see zarr files in sirius_data/telescope_layout/data/).
    parallactic_angles: float np.array, radians
        An array of the parallactic angles for which to calculate the antenna beams.
    freq_chan: float np.array, [n_chan], Hz
        Channel frequencies.
    beam_parms: dict
    beam_parms['mueller_selection']: int np.array, default=np.array([ 0, 5, 10, 15])
        The elements in the 4x4 beam Mueller matrix to use. The elements are numbered row wise.
        For example [ 0, 5, 10, 15] are the diagonal elements.
    beam_parms['pa_radius']: float, default=0.2, radians
        The change in parallactic angle that will trigger the calculation of a new beam when using Zernike polynomial aperture models.
    beam_parms['image_size']: int np.array, default=np.array([1000,1000])
        Size of the beam image generated from the Zernike polynomial coefficients.
    beam_parms['fov_scaling']: int, default=1.2
        This parameter should rarely be modified. Used to determine the cell size of the beam image so that it lies within the image that is generated.
    beam_parms['zernike_freq_interp']: str, default='nearest', options=['linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic']
        What interpolation method to use for Zernike polynomial coefficients.
        
    Returns
    -------
    J_xds: xr.Dataset
        An xds that contains the image of the per antenna beam as a function of [pa (paralactic angle), chan (channel), pol (polarization), l (orthographic/synthesis projection of directional cosine), m (orthographic/synthesis projection of directional cosine)]. Should not be confused with the primary beam, which is the beam for a baseline and is equal to the product of two antenna beams.
    """
    
    #Calculate parallactic angles.
    pa = _calc_parallactic_angles_astropy(time_str,np.array([site_location['m0']['value'],site_location['m1']['value'],site_location['m2']['value']]),phase_center_ra_dec)
    #print('pa',pa*180/np.pi)
    pa_subset,_,_ = _find_optimal_set_angle(pa,beam_parms['pa_radius'] )
    #print('pa_subset',pa_subset*180/np.pi)
    
    _beam_parms = copy.deepcopy(beam_parms)

    # If beam model is a Zernike polynomial coefficient xr.Datasets convert it to an image.
    eval_beam_models = []
    for bm in beam_models:
        if 'ZPC' in bm: #check for zpc files
            J_xds = calc_zpc_beam(bm,pa_subset,freq_chan,_beam_parms,check_parms)
            J_xds.attrs = bm.attrs
            eval_beam_models.append(J_xds)
        else:
            eval_beam_models.append(bm)
    
    return eval_beam_models, pa


def make_mueler_mat(J_xds1, J_xds2, mueller_selection):
    #print(J_xds1.pol,J_xds1.pol.values)
    pol_indx = _pol_code_to_index(J_xds1.pol.values)
    pol1 = J_xds1.pol.values
    pol2 = J_xds2.pol.values
    #pa,chan,l,m must match
    J1_shape = J_xds1.J.shape
    J2_shape = J_xds2.J.shape
    
    #print(J1_shape,J2_shape)
    
    M_shape = J1_shape[0:2] + (len(mueller_selection),) + J1_shape[3:5]
    M = np.zeros(M_shape,np.complex)
    m_sel = mueller_selection
    pol1=np.zeros(len(mueller_selection))
    pol2=np.zeros(len(mueller_selection))

    for i,m_flat_indx in enumerate(mueller_selection):
        #print(m_flat_indx,m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4))
        #print(map_mueler_to_pol[m_flat_indx,0],map_mueler_to_pol[m_flat_indx,1])
        #print(np.where(pol_indx == map_mueler_to_pol[m_flat_indx,0])[0][0],np.where(pol_indx == map_mueler_to_pol[m_flat_indx,1])[0][0])
        M[:,:,i,:,:] = J_xds1.J[:,:,np.where(pol_indx == map_mueler_to_pol[m_flat_indx,0])[0][0],:,:]*np.conj(J_xds2.J[:,:,np.where(pol_indx == map_mueler_to_pol[m_flat_indx,1])[0][0],:,:])
        pol1[i]=_index_to_pol_code(map_mueler_to_pol[m_flat_indx,0],J_xds1.pol.values)
        pol2[i]=_index_to_pol_code(map_mueler_to_pol[m_flat_indx,1],J_xds2.pol.values)
    '''
    M_shape = J1_shape[0:2] + (4,4) + J1_shape[3:5]
    M = np.zeros(M_shape,np.complex)

    for m_flat_indx in mueller_selection:
        #print(m_flat_indx,m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4))
        #print(map_mueler_to_pol[m_flat_indx,0],map_mueler_to_pol[m_flat_indx,1])
        #print(np.where(pol_indx == map_mueler_to_pol[m_flat_indx,0])[0][0],np.where(pol_indx == map_mueler_to_pol[m_flat_indx,1])[0][0])
        M[:,:,m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4),:,:] = J_xds1.J[:,:,np.where(pol_indx == map_mueler_to_pol[m_flat_indx,0])[0][0],:,:]*np.conj(J_xds2.J[:,:,np.where(pol_indx == map_mueler_to_pol[m_flat_indx,1])[0][0],:,:])
    '''
        
    l = J_xds1.l
    m = J_xds1.m
    freq_chan = J_xds1.chan
    parallactic_angles = J_xds1.pa
    
    coords = {'chan':freq_chan, 'pa': parallactic_angles,'m_sel': m_sel,'pol1':('m_sel',pol1),'pol2':('m_sel',pol2),'l':l,'m':m}
    
    M_xds = xr.Dataset()
    M_xds = M_xds.assign_coords(coords)
    
    M_xds['M'] = xr.DataArray(M, dims=['pa','chan','m_sel','l','m'])
    
        
            
    return M_xds
    



#  CASA Next Generation Infrastructure
#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.



### Gaussian Beam Source Model Functions ###
# To understand this math see memo https://drive.google.com/file/d/1ngPiUsGDazhfno8RxV4MXuPI6SnbLgO0/view

##import numpy as np
from scipy.fftpack import fft2, ifft2, fftshift

#Performs convolution using fft
def _fft_conv(x,y,pix):
    return pix*pix*np.real(fftshift(ifft2((fft2(x))*(fft2(y)))))

#Convert Gaussian beam paramters from d1, d2, t (semi-major, semi-minor, theta) to a, b, g (alpha, beta, gamma)
def _convert_to_abg(parms_ddt):
    d1 = parms_ddt[0]
    d2 = parms_ddt[1]
    t = parms_ddt[2]
    m = (4*np.log(2))/(d1**2)
    n = (4*np.log(2))/(d2**2)
    a = m*(np.cos(t)**2) + n*(np.sin(t)**2)
    b = 2*(m-n)*np.sin(t)*np.cos(t)
    g = m*(np.sin(t)**2) + n*(np.cos(t)**2)
    return np.array([a, b, g])

#Convert Gaussian beam paramters from a, b, g (alpha, beta, gamma) to d1, d2, t (semi-major, semi-minor, theta)
def _convert_to_ddt(parms_abg):
    a = parms_abg[0]
    b = parms_abg[1]
    g = parms_abg[2]
    d1 = np.sqrt((8*np.log(2))/(a + g - np.sqrt(a**2 - 2*a*g + g**2 + b**2)))
    d2 = np.sqrt((8*np.log(2))/(a + g + np.sqrt(a**2 - 2*a*g + g**2 + b**2)))
    t = 0.5*np.arctan2(-b,g-a)
    return np.array([d1, d2, t])

#Calculates the abg parameters for a Fourier Transformed Gaussian
def _fft_parms_abg(parms_abg):
    a = parms_abg[0]
    b = parms_abg[1]
    g = parms_abg[2]
    a_ft = (4*a*g*np.pi**2)/(4*(a**2)*g-a*(b**2))
    print(a_ft)
    b_ft = (-4*a*b*np.pi**2)/(4*(a**2)*g-a*(b**2))
    g_ft = (4*(a**2)*np.pi**2)/(4*(a**2)*g-a*(b**2))
    return np.array([a_ft, b_ft, g_ft])

#Find the amplitude and ddt parameters for the correcting beam
def _gaussian_deconvolve(new_beam_parms_ddt,old_beam_parms_ddt,A_new_beam,A_old_beam):
    nb_abg = convert_to_abg(new_beam_parms_ddt)
    ob_abg = convert_to_abg(old_beam_parms_ddt)
    ft_nb_abg = fft_parms_abg(nb_abg)
    ft_ob_abg = fft_parms_abg(ob_abg)
    corrected_parms_abg = fft_parms_abg(ft_nb_abg-ft_ob_abg)
    corrected_parms_ddt = convert_to_ddt(corrected_parms_abg)
    
    A_nb_ft = ft_amp(new_beam_parms_ddt, A_new_beam)
    A_ob_ft = ft_amp(old_beam_parms_ddt, A_old_beam)
    A_cor = ift_amp(corrected_parms_ddt, A_nb_ft/A_ob_ft)
    
    return  A_cor, corrected_parms_ddt

#Fourier transformed amplitude of beam
def _ft_amp(beam_parms_ddt, A):
    return A*(np.pi*beam_parms_ddt[0]*beam_parms_ddt[1])/np.log(16)

#Inverse Fourier transformed amplitude of beam
def _ift_amp(beam_parms_ddt, A):
    return A*np.log(16)/(np.pi*beam_parms_ddt[0]*beam_parms_ddt[1])
    
#Generating beams from parameters
def _gauss_beam_ddt(amp,parms_ddt,x,y):
    '''
    parms_ddt:
    
    '''
    d1 = parms_ddt[0] #semi major axis
    d2 = parms_ddt[1] #semi minor axis
    t = parms_dtt[2] # positionangle t, is the anti-clockwise angle from the x axis to the line that lies along the greatest width
    x_r = np.cos(t)*x + np.sin(t)*y
    y_r = -np.sin(t)*x + np.cos(t)*y
    r = 4*np.log(2)*(x_r**2)/(d1**2) + 4*np.log(2)*(y_r**2)/(d2**2)
    
    B = amp*np.exp(-r)
    #B[r > 20] = 0.0
    return B

def _gauss_beam_abg(amp,parms_abg,x,y):
    a = parms_abg[0]
    b = parms_abg[1]
    g = parms_abg[2]
    r = a*(x**2) + b*x*y + g*y**2
    B = amp*np.exp(-r)
    #B[r > 20] = 0
    return B




import os

import numpy as np
import pkg_resources
import pytest
import xarray as xr
from astropy.coordinates import SkyCoord

from sirius import simulation
from sirius.dio import make_chan_xda, make_time_xda
from sirius_data.beam_1d_func_models.airy_disk import vla


@pytest.fixture()
def source_radec_template():
    source_skycoord = SkyCoord(ra="19h59m50.51793355s", dec="+40d48m11.3694551s", frame="fk5")
    source_ra_dec = np.array([source_skycoord.ra.rad, source_skycoord.dec.rad])[None, None, :]
    return source_ra_dec


@pytest.fixture()
def phase_center_radec_template():
    phase_center = SkyCoord(ra="19h59m28.5s", dec="+40d44m01.5s", frame="fk5")
    phase_center_radec = np.array([phase_center.ra.rad, phase_center.dec.rad])[None, :]
    return phase_center_radec


@pytest.fixture()
def tel_xds_template():
    tel_dir = pkg_resources.resource_filename("sirius_data", "telescope_layout/data/vla.d.tel.zarr")
    tel_xds = xr.open_zarr(tel_dir, consolidated=False)
    return tel_xds


@pytest.fixture()
def time_xda_template():
    time_xda = make_time_xda(
        time_start="2019-10-03T19:00:00.000",
        time_delta=3600,
        n_samples=10,
        n_chunks=1,
    )
    return time_xda


@pytest.fixture()
def chan_xda_template():
    chan_xda = make_chan_xda(
        freq_start=3 * 10**9,
        freq_delta=0.4 * 10**9,
        freq_resolution=0.01 * 10**9,
        n_channels=3,
        n_chunks=1,
    )
    return chan_xda


@pytest.mark.skip(reason="xfail due to https://github.com/casangi/sirius/issues/2 but skipping is faster w/ fixtures")
@pytest.mark.parametrize("to_disk", [(False, True)])
def test_simple_simulation(
    source_radec_template,
    phase_center_radec_template,
    tel_xds_template,
    time_xda_template,
    chan_xda_template,
    to_disk,
):
    """
    Adapted from docs/simple_simulation.ipynb
    """
    xds = simulation(
        # source flux
        np.array([1.0, 0, 0, 1.0])[None, None, None, :],
        source_radec_template,
        # pointing_radec
        None,
        phase_center_radec_template,
        # phase center names
        np.array(["field1"]),
        # beam_params
        dict(),
        # beam_models
        [vla],
        # beam_model_map
        np.zeros(27, dtype=int),
        # uvw_params
        {"calc_method": "casa", "auto_corr": False},
        tel_xds_template,
        time_xda_template,
        chan_xda_template,
        # pol
        [5, 8],
        # noise_params
        None,
        # save_params
        {
            "ms_name": "simple_sim.ms",
            "write_to_ms": bool(to_disk),
        },
    )
    xds.compute()
    assert os.path.exists("simple_sim.ms")



 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import warnings, time, os, psutil, multiprocessing, re
import dask
import copy
from ._parm_utils._check_logger_parms import _check_logger_parms, _check_worker_logger_parms
from sirius._sirius_utils._sirius_logger import _sirius_worker_logger_plugin, _setup_sirius_logger

def initialize_processing(cores=None, memory_limit=None, log_parms={}, worker_log_parms={}):
    '''
    https://github.com/dask/dask/issues/5577
    log_parms['log_to_term'] = True/False
    log_parms['log_file'] = True/False
    log_parms['level'] =
    
    '''
    _log_parms = copy.deepcopy(log_parms)
    _worker_log_parms = copy.deepcopy(worker_log_parms)
    
    assert(_check_logger_parms(_log_parms)), "######### ERROR: initialize_processing log_parms checking failed."
    assert(_check_worker_logger_parms(_worker_log_parms)), "######### ERROR: initialize_processing log_parms checking failed."

    # setup dask.distributed based multiprocessing environment
    if cores is None: cores = multiprocessing.cpu_count()
    if memory_limit is None: memory_limit = str(round(((psutil.virtual_memory().available / (1024 ** 2)) * 0.90) / cores)) + 'MB'
    dask.config.set({"distributed.scheduler.allowed-failures": 10})
    dask.config.set({"distributed.scheduler.work-stealing": False})
    dask.config.set({"distributed.scheduler.unknown-task-duration": '99m'})
    dask.config.set({"distributed.worker.memory.pause": False})
    dask.config.set({"distributed.worker.memory.terminate": False})
    #dask.config.set({"distributed.worker.memory.recent-to-old-time": '999s'})
    dask.config.set({"distributed.comm.timeouts.connect": '3600s'})
    dask.config.set({"distributed.comm.timeouts.tcp": '3600s'})
    dask.config.set({"distributed.nanny.environ.OMP_NUM_THREADS": 1})
    dask.config.set({"distributed.nanny.environ.MKL_NUM_THREADS": 1})
    cluster = dask.distributed.LocalCluster(n_workers=cores, threads_per_worker=1, processes=True, memory_limit=memory_limit) #, silence_logs=logging.ERROR
    #cluster = dask.distributed.LocalCluster(n_workers=1, threads_per_worker=24, processes=True, memory_limit='512GB')
    client = dask.distributed.Client(cluster)
    
    client.get_versions(check=True)
    from distributed.plugins.autorestrictor import install_plugin
    client.run_on_scheduler(install_plugin)
    
    import time
    start = time.time()
    #client.run(init_logging)
    #client.register_worker_callbacks()
    
    _setup_sirius_logger(log_to_term=_log_parms['log_to_term'],log_to_file=_log_parms['log_to_file'],log_file=_log_parms['log_file'], level=_log_parms['log_level'])
    #print('main logger time ', time.time()-start)
    
    
    #start = time.time()
    worker_logger = _sirius_worker_logger_plugin(_worker_log_parms)
    #print('2worker loggers time ', time.time()-start)
    
    #start = time.time()
    client.register_worker_plugin(plugin=worker_logger, name='sirius_worker_logger')
    #print('3worker loggers time ', time.time()-start)
    
    return client



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np

def _get_time_ha_aligned_time_casa(tel_xds,start_date,phase_center_ra_dec):
    from casatools import simulator, measures
    from sirius._sirius_utils._cngi_io import read_ms
    me = measures()
    import os
    import numpy as np

    sm = simulator()
    ant_pos = tel_xds.ANT_POS.values
    os.system("rm -rf temp_42.ms")
    sm.open(ms="temp_42.ms")

    ## Set the antenna configuration
    sm.setconfig(
        telescopename=tel_xds.telescope_name,
        x=ant_pos[:, 0],
        y=ant_pos[:, 1],
        z=ant_pos[:, 2],
        dishdiameter=tel_xds.DISH_DIAMETER.values,
        mount=["alt-az"],
        antname=list(
            tel_xds.ant_name.values
        ),  # CASA can't handle an array of antenna names.
        coordsystem="global",
        referencelocation=tel_xds.site_pos[0],
    )

    sm.setfeed(mode="perfect X Y", pol=[""])


    sm.setspwindow(
        spwname="temp",
        freq="1GHz",
        deltafreq="0.5GHz",
        freqresolution="0.5GHz",
        nchannels=1,
        refcode="LSRK",
    )


    sm.setauto(autocorrwt=0.0)

    sm.settimes(
        integrationtime='60s',
        usehourangle=True,
        referencetime=me.epoch('utc', start_date),
    )
    
    
    dir_dict = {"m0": {"unit": "rad", "value": phase_center_ra_dec[0]},
                "m1": {"unit": "rad", "value": phase_center_ra_dec[1]},
                "refer": "J2000",
                "type": "direction",}
    
    sm.setfield(sourcename="temp_field", sourcedirection=dir_dict)


    sm.observe(sourcename="temp_field",
                spwname="temp",
                starttime="0h",
                stoptime="1h",)

    sm.close()
    
    mxds = read_ms("temp_42.ms")
    start_time = mxds.xds0.TIME.values[0]
    
    return start_time
    
    
    
    
    
def _get_time_ha_aligned_time(earth_location,initial_start_time,phase_center_ra_dec):
    import casatools
    qa = casatools.quanta()
    qq = qa.quantity

    ref = 'LAST'
    me = casatools.measures()
    me.doframe(earth_location)
    last = me.measure(me.epoch('utc', initial_start_time), ref) #Convert utc to Local Apparent Sidereal Time
    new_last = np.round(last['m0']['value']) + phase_center_ra_dec[0]/(2*np.pi) # LAST = HA + RA
    new_last = {'m0': {'value': new_last, 'unit': 'd'},'refer': ref,'type': 'epoch'}
    new_start_time = me.measure(new_last, 'UTC')
    new_start_time = qa.time({'value': new_start_time['m0']['value'], 'unit': 'd'},form="fits")[0]
    
    return(new_start_time)



 #   Copyright 2019 AUI, Inc. Washington DC, USA
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
 #   you may not use this file except in compliance with the License.
 #   You may obtain a copy of the License at
 #
 #       http://www.apache.org/licenses/LICENSE-2.0
 #
 #   Unless required by applicable law or agreed to in writing, software
 #   distributed under the License is distributed on an "AS IS" BASIS,
 #   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #   See the License for the specific language governing permissions and
 #   limitations under the License.
 
import numpy as np
import itertools
import time
import xarray as xr
import dask.array as da
import numba
from numba import jit
from scipy import signal
from sirius._sirius_utils._beam_funcs import _casa_airy_beam, _airy_beam, _casa_airy_beam_njit, _airy_beam_njit, _poly_beam
from sirius._sirius_utils._math_utils import _interp_array
from sirius._sirius_utils._coord_transforms import  _rot_coord, _compute_rot_coords
from sirius._sirius_utils._zernike_polynomials import _generate_zernike_surface
from sirius._sirius_utils._array_utils import _find_angle_indx, _find_val_indx
from sirius_data._constants import map_mueler_to_pol, c


def _calc_resolution(beam_freq,dish_diameter,beam_parms):
    # Ensures that the beam lies within the image.
    fov = beam_parms['fov_scaling']*(1.22 * c / (dish_diameter*beam_freq))
    
    print(fov)
    max_delta = max(max(fov/beam_parms['image_size'][0]),max(fov/beam_parms['image_size'][1]))
    return max_delta

def _beam_models_to_tuple(beam_models):
    '''
    Convert different beam models into a single format that consists of tupples and arrays.
    This is required by numba that does not support dictionaries.
    
    The function changes beam_models to a tuple where each element is tuple containing a beam model (tuple of tuples).
    The elements in the beam model tuple are strings, numbers and arrays describings the beam model.
    
    The beam model format:
    ( type: int
        0-image based beam model
        1-function based beam model
      func: str
        'airy'
        'casa_airy'
        'none'-image based
      J/par: complex np.array, [n_pa, n_chan, n_pol, n_l, n_m]/[1,1,1,1,n_par]
        The image or parameters for the beam function.
      pa: float np.array, [n_pa]
        Parallactic angles.
      chan: float np.array [n_chan]
        Channel frequencies.
      pol: int np.array [n_pol]
        Pol index (not pol code).
      delta_l: float
      delta_m: float
      max_rad_1GHz: float )
    '''
    

    beam_models_list = []
    
    for i,bm in enumerate(beam_models):
        if "J" in bm: #Image based beam
            bm_tuple = _exstract_arrays_from_img_bm_xds(bm)
            beam_models_list.append(bm_tuple)
        elif "func" in bm:
            bm_tuple = _exstract_vals_from_analytic_dict(bm)
            beam_models_list.append(bm_tuple)
        elif "BPC" in bm:
            bm_tuple = _exstract_arrays_from_poli_bm_xds(bm)
            beam_models_list.append(bm_tuple)
        else:
            assert False, print('Unsupported beam model:',bm)
        
    return tuple(beam_models_list)

    #(type: int, func: str,J/par: complex np.array, [n_pa, n_chan, n_pol, n_l, n_m]/[1,1,1,1,n_par], pa: float np.array, [n_pa], chan: float np.array [n_chan], pol: int np.array [n_pol], delta_l: float, delta_m: float ,max_rad_1GHz: float)
    
    

    
def check_bm(bm):
    '''
    Ensure bm is in correct format otherwise numba code will malfunction.
    '''
    assert isinstance(bm[0],np.int)
    assert isinstance(bm[1],str)
    assert isinstance(bm[2],np.ndarray)
    assert isinstance(bm[2][0,0,0,0,0],complex)
    assert len(bm[2].shape)==5
    assert isinstance(bm[3],np.ndarray)
    assert isinstance(bm[3][0],float)
    assert len(bm[3].shape)==1
    assert isinstance(bm[4],np.ndarray)
    assert isinstance(bm[4][0],float)
    assert len(bm[4].shape)==1
    assert isinstance(bm[5],np.ndarray)
    assert isinstance(bm[5][0],np.int64)
    assert len(bm[5].shape)==1
    assert isinstance(bm[6],float)
    assert isinstance(bm[7],float)
    assert isinstance(bm[8],float)
    
def _exstract_arrays_from_poli_bm_xds(bm):
    #(type: int, func: str,J: complex np.array, [n_pa, n_chan, n_pol, n_l, n_m], pa: float np.array, [n_pa], chan: float np.array [n_chan], pol: int np.array [n_pol], delta_l: float, delta_m: float ,max_rad_1GHz: float)
    bm_BPC = bm.BPC.values.astype(np.complex128)[None,:,:,None,:]
    #print("bm_J", bm_J.shape)
    chan = bm.chan.values
    pol = _pol_code_to_index(bm.pol.values)
    max_rad_1GHz = bm.attrs['max_rad_1GHz']
    func = 'poli'
    f_arr = np.array([0.0])
    i_arr = np.array([0])
    return (2,func,bm_BPC,f_arr,chan,pol,0.0,0.0,max_rad_1GHz)

def _exstract_arrays_from_img_bm_xds(bm):
    #(type: int, func: str,J: complex np.array, [n_pa, n_chan, n_pol, n_l, n_m], pa: float np.array, [n_pa], chan: float np.array [n_chan], pol: int np.array [n_pol], delta_l: float, delta_m: float ,max_rad_1GHz: float)
    bm_J = bm.J.values 
    #print("bm_J", bm_J.shape)
    pa = bm.pa.values
    chan = bm.chan.values
    pol = _pol_code_to_index(bm.pol.values)
    delta_l = (bm.l[1].values - bm.l[0].values).astype(float)
    delta_m = (bm.m[1].values - bm.m[0].values).astype(float)
    max_rad_1GHz = bm.attrs['max_rad_1GHz']
    func = 'image'
    return (0,func,bm_J,pa,chan,pol,delta_l,delta_m,max_rad_1GHz)
    
def _exstract_vals_from_analytic_dict(bm):
    #(type: int, func: str,J/par: complex np.array, [1,1,1,1,n_par], -: float np.array, [1], -: float np.array [1], -: int np.array [1], -: float, -: float,max_rad_1GHz: float)
    func = bm['func']
    #dish_diameter = bm['dish_diameter']
    #blockage_diameter = bm['blockage_diameter']
    dish_diameter = bm['dish_diam']
    blockage_diameter = bm['blockage_diam']
    max_rad_1GHz = bm['max_rad_1GHz']
    
    par_arr = np.zeros((1,1,1,1,2),np.complex)
    par_arr[0,0,0,0,0] = dish_diameter
    par_arr[0,0,0,0,1] = blockage_diameter
    f_arr = np.array([0.0])
    i_arr = np.array([0])
    return (1,func,par_arr,f_arr,f_arr,i_arr,0.0,0.0,max_rad_1GHz)

#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _calc_pb_scale(flux, sep1, sep2,lmn1,lmn2,bm1_indx,bm2_indx,bm1,bm2,pa,freq,mueller_selection,do_pointing):
    # bm tuple: (type: int, func: str,J/par: complex np.array, [n_pa, n_chan, n_pol, n_l, n_m]/[1,1,1,1,n_par], pa: float np.array, [n_pa], chan: float np.array [n_chan], pol: int np.array [n_pol], delta_l: float, delta_m: float ,max_rad_1GHz: float)
    # bm[0]: 0 (zernike), 1 (func), 2 (polinomial)
    outside_beam = False
    max_rad_1 = bm1[8]/(freq/10**9) # scale max_rad_1GHz (bm[8]) to freq
    max_rad_2 = bm2[8]/(freq/10**9) # scale max_rad_1GHz (bm[8]) to freq
    
    #Create Jones matrix J for first antenna
    if (sep1 < max_rad_1) and (sep2 < max_rad_2):
        if bm1[0] == 0: #Image beam model
            J_sampled_1 = _sample_J_image(bm1[2],bm1[3],bm1[4],bm1[5],bm1[6],bm1[7],pa,freq,lmn1)
        elif bm1[0] == 1: #Function based beam model
            J_sampled_1 = _sample_J_func(bm1[1],bm1[2][0,0,0,0,:],bm1[8],lmn1,freq,1)
        elif bm1[0] == 2: #Polynomial based model
            J_sampled_1 = _sample_J_poli(bm_BPC=bm1[2][0,:,:,0,:],bm_chan=bm1[4],bm_pol=bm1[5],pa=pa,freq=freq,lmn=lmn1,ipower=1,max_rad_1GHz=bm1[8])
    else:
        outside_beam = True
        flux_scaled = np.zeros(4, dtype = numba.complex128)
        return flux_scaled, outside_beam
        
    #Create Jones matrix J for second antenna if different from first.
    if (bm1_indx == bm2_indx) and ~do_pointing:
        J_sampled_2 = J_sampled_1
    else:
        if bm2[0] == 0: #Image beam model
            J_sampled_2 = _sample_J_image(bm2[2],bm2[3],bm2[4],bm2[5],bm2[6],bm2[7],pa,freq,lmn2)
        elif bm2[0] == 1: #Function based beam model
            J_sampled_2 = _sample_J_func(bm2[1],bm2[2][0,0,0,0,:],bm2[8],lmn2,freq,1)
        elif bm2[0] == 2: #Polynomial based model
            J_sampled_2 = _sample_J_poli(bm_BPC=bm2[2][0,:,:,0,:],bm_chan=bm2[4],bm_pol=bm2[5],pa=pa,freq=freq,lmn=lmn2,ipower=1,max_rad_1GHz=bm2[8])
    
    if (bm1[0] in [1,2]) and (bm2[0] in [1,2]): #The function and polynomial coefficient based beam models have no leakage terms.
        flux_scaled = flux*(J_sampled_1*J_sampled_2)
    else:
        M = _make_mueler_mat(J_sampled_1, J_sampled_2, mueller_selection)
        flux_scaled = np.dot(M,flux)
        
    return flux_scaled, outside_beam

@jit(nopython=True,cache=True,nogil=True)
def _pol_code_to_index(pol):
    if pol[0] in [5,6,7,8]:
        return pol-5
    if pol[0] in [9,10,11,12]:
        return pol-9
    if pol[0] in [0,1,2,3]:
        return pol
    assert False # "Unsupported pol " + str(pol) #Numba does not support an explicit error message.
    
@jit(nopython=True,cache=True,nogil=True)
def _index_to_pol_code(index,pol):
    if pol[0] in [5,6,7,8]:
        return index+5
    if pol[0] in [9,10,11,12]:
        return index+9
    assert False # "Unsupported pol " + str(pol) #Numba does not support an explicit error message.

    
    
#@jit(nopython=True,cache=True,nogil=True)
@jit(nopython=True,nogil=True)
def _sample_J_func(func, parms,max_rad_1GHz, lmn, freq, ipower):
    #beam_parms = bm
    #beam_parms['ipower'] = 1
    
    if func == 'casa_airy':
        J_sampled = _casa_airy_beam(lmn[0],lmn[1],freq,parms[0].real, parms[1].real,ipower,max_rad_1GHz)
        #J_sampled = 0.5
    elif func == 'airy':
        #J_sampled = 0.5
        J_sampled = _airy_beam(lmn[0],lmn[1],freq,parms[0].real,parms[1].real, ipower)
    else:
        J_sampled = 1
    
    J_sampled_array = np.zeros(4, dtype = numba.float64)
    J_sampled_array[0] = J_sampled
    J_sampled_array[3] = J_sampled
    #J_sampled = np.array([J_sampled,0,0,J_sampled])
    return J_sampled_array.astype(numba.complex128)
    
@jit(nopython=True,cache=True,nogil=True)
def _sample_J_image(bm_J,bm_pa,bm_chan, bm_pol, bm_delta_l,bm_delta_m,pa,freq,lmn):
    """
    Obtain Jones matrix at coordinates lmn.
    Dimensionality refers to the beam model (not visibilities).
    
    Parameters
    ----------
    bm_J: complex np.array, [n_pa, n_chan, n_pol, n_pixels_x, n_pixels_y]
        Input array of Jones matrices for various parallactic angles, channels, and polarizations. 
    bm_pa: int np.array, [n_pa], radians
        Array of available parallactic angle indices
    bm_chan: int np.array, [n_chan], Hertz
        Array of available frequency channel indices
    bm_pol: int np.array, [n_pol]
        Array of available Jones matrix indices (valid values are 0,1,2,3. Jones matrix is flattened row-wise.)
    bm_delta_l: float, radians/pixel
        Radians per pixel along x
    bm_delta_m: float, radians/pixel
        Radians per pixel along y
    pa: float, radians
        Parallactic angle of desired Jones matrix
    freq:  float, Hertz
        Value of frequency of desired Jones matrix
    lmn: float np.array, [3], radians
        l,m,n coordinates of desired Jones matrix 
    
    Returns
    -------
    J: float np.array, [4]
        Interpolated Jones matrix at desired parallactic angle and coordinates
    
    """
    pa_indx = _find_angle_indx(bm_pa,pa)
    chan_indx = _find_val_indx(bm_chan, freq)
    bm_J_sub = bm_J[pa_indx, chan_indx,:,:,:]
    
    freq_scaling_factor = freq/bm_chan[chan_indx] # freq of visibility/freq of beam model
    x_rot, y_rot  = _rot_coord(lmn[0]*freq_scaling_factor,lmn[1]*freq_scaling_factor,pa-bm_pa[pa_indx])
    #xrot = np.ones(1, dtype = numba.float64)
    xrot = np.ones(1, dtype = np.float64)
    xrot[0] = x_rot
    #yrot = np.ones(1, dtype = numba.float64)
    yrot = np.ones(1, dtype = np.float64)
    yrot[0] = y_rot
    #print((xrot/bm[4]) + len(bm_J_sub[0, :, 0])//2)
    #print((yrot/bm[5]) + len(bm_J_sub[0, 0, :])//2)
    J_temp = _interp_array(bm_J_sub, xrot, yrot, bm_delta_l, bm_delta_m)
    
    J = np.zeros((4,),dtype=numba.complex128)
    for i,p in enumerate(bm_pol):
        J[p] = J_temp[i,0]
        #print(i)
        
    return J

@jit(nopython=True,cache=True,nogil=True)
def _sample_J_poli(bm_BPC,bm_chan,bm_pol,pa,freq,lmn,ipower,max_rad_1GHz):
    '''
    bm_BPC: float np.array, [n_chan,n_pol,n_coef]
    '''
    
    chan_indx = _find_val_indx(bm_chan, freq)
    bm_BPC_sub = bm_BPC[chan_indx,:,:]
    
    #freq_scaling_factor = freq/bm_chan[chan_indx] # freq of visibility/freq of beam model
    #J_sampled = _poly_beam(lmn[0]*freq_scaling_factor,lmn[1]*freq_scaling_factor,bm_chan[chan_indx],bm_BPC_sub.real, ipower,max_rad_1GHz)
    J_sampled = _poly_beam(lmn[0],lmn[1],freq,bm_BPC_sub.real, ipower,max_rad_1GHz)
    
    J_sampled_array = np.zeros(4, dtype = numba.float64)
    J_sampled_array[0] = J_sampled
    J_sampled_array[3] = J_sampled
    #J_sampled = np.array([J_sampled,0,0,J_sampled])
    
    #print('In sample_J',J_sampled,bm_chan[chan_indx],freq,lmn)
    
    return J_sampled_array.astype(numba.complex128)

#Non-numba versions:
def _sample_J_func_og(bm,lmn,freq):
    beam_parms = bm
    beam_parms['ipower'] = 1
    
    if beam_parms['func'] == 'casa_airy':
        J_sampled = _casa_airy_beam_njit(lmn,freq,beam_parms)
    elif beam_parms['func'] == 'airy':
        J_sampled = _airy_beam_njit(lmn,freq,beam_parms)
    else:
        J_sampled = 1
    J_sampled = np.array([J_sampled,0,0,J_sampled])
    return J_sampled


def _sample_J_image_og(bm,lmn,freq,pa):
# sample_J(J,pa_,chan_,lmn,chan,pa)
    bm_sub = bm
    if len(bm.pa) > 1:
        pa_indx= _find_angle_indx(bm_sub.pa.values,pa)
        bm_sub = bm_sub.isel(pa=pa_indx)
    else:
        bm_sub = bm_sub.isel(pa=0)
    if len(bm.chan) > 1:
        bm_sub = bm.J.interp(chan=freq,method='nearest')
    else:
        bm_sub = bm_sub.isel(chan=0)
        
    #print('pa values',bm_sub)
    x_rot, y_rot  = _rot_coord(lmn[0],lmn[1],pa-bm_sub.pa.values)
    
#    print(bm_sub)
#    print(lmn,pa,bm_sub.pa.values )
#    print(x_rot,y_rot)
#    print(bm_sub.J.isel(model=0).interp(l=x_rot,m=y_rot,method='linear'))
#    print(bm_sub.J.interp(l=x_rot,m=y_rot,method='linear').values)
    
    return bm_sub.J.interp(l=x_rot,m=y_rot,method='linear').values[0]

@jit(nopython=True,cache=True,nogil=True)
def _make_mueler_mat(J1, J2, mueller_selection):

    #M = np.zeros((4,4),dtype=np.complex)
    M = np.zeros((4,4),dtype=numba.complex128)
    
    for m_flat_indx in mueller_selection:
        #print(m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4))
        #print(np.where(map_mueler_to_pol[m_flat_indx][0] == pol)[0][0])
        #print(pol, map_mueler_to_pol[m_flat_indx][0])
        #M[m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4)] = J1[np.where(map_mueler_to_pol[m_flat_indx][0] == pol)[0][0]]*np.conj(J2[np.where(map_mueler_to_pol[m_flat_indx][1] == pol)[0][0]])
        
        M[m_flat_indx//4,m_flat_indx - 4*(m_flat_indx//4)] = J1[map_mueler_to_pol[m_flat_indx,0]]*np.conj(J2[map_mueler_to_pol[m_flat_indx,1]])
            
    return M
    
    #map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])

'''
    #print('pol',pol)
    if inv:
        map_mueler_to_pol = np.array([[3, 3],[3, 2],[2, 3],[2, 2],[3, 1],[3, 0],[2, 1],[2, 0],[1, 3],[1, 2],[0, 3],[0, 2],[1, 1],[1, 0],[0, 1],[0, 0]]) # np.flip(map_mueler_to_pol,axis=0)
        #map_mueler_to_pol = np.array([ [[3, 3],[3, 2],[2, 3],[2, 2]],[[3, 1],[3, 0],[2, 1],[2, 0]],[[1, 3],[1, 2],[0, 3],[0, 2]],[[1, 1],[1, 0],[0, 1],[0, 0]]])
    else:
        map_mueler_to_pol = np.array([[0,0],[0,1],[1,0],[1,1],[0,2],[0,3],[1,2],[1,3],[2,0],[2,1],[3,0],[3,1],[2,2],[2,3],[3,2],[3,3]])
        #map_mueler_to_pol = np.array([[[0,0],[0,1],[1,0],[1,1]],[[0,2],[0,3],[1,2],[1,3]],[[2,0],[2,1],[3,0],[3,1]],[[2,2],[2,3],[3,2],[3,3]]])
        
'''

# Do we need the abs for apeture_parms['zernike_size'].
#    apeture_parms['cell_size'] = 1/(beam_parms['cell_size']*beam_parms['image_size']*apeture_parms['oversampling'])
#    apeture_parms['zernike_size'] = np.floor(np.abs(Dish_Diameter*eta/(apeture_parms['cell_size']*lmbd))) #Why is abs used?
#2.) Currently in ARD-20 the Zernike grid parameters are caculated using (in ZernikeCalc.cc):
#size_zernike = floor((D*eta)/dx)
#delta_zernike = 2.0/size_zernike
#
#By reversing the order the delta of the zernike grid and uv grid can match exactly (no floor operation). The zernike grid will be slightly larger (all cells outside the r=1 will still be nulled).
#delta_zernike = (2.0*dx)/(D*eta)
#size_zernike = ceil((D*eta)/dx)
#Assume ETA is the same for all pol and coef. Only a function of freq. Need to update zpc.zarr format.
  
def _calc_ant_jones(zpc_dataset,j_freq,j_pa,beam_parms):
    pa_prev = -42.0
    freq_prev = -42.0
    
    j_planes = np.zeros((len(j_pa),len(j_freq),len(beam_parms['needed_pol']),beam_parms['image_size'][0],beam_parms['image_size'][1]),np.complex128)
    j_planes_shape = j_planes.shape
    iter_dims_indx = itertools.product(np.arange(j_planes_shape[0]),np.arange(j_planes_shape[1]))
    ic = beam_parms['image_size']//2 #image center pixel
    
    
    #print('j_planes_shape,ic',j_planes_shape,ic)
    
    for i_pa, i_chan in iter_dims_indx:
        #print(i_pa,i_chan)
        pa = j_pa[i_pa]
        beam = zpc_dataset
        freq = j_freq[i_chan]

        if(freq != freq_prev):
            beam_interp = beam.interp(chan=freq,method=beam_parms['zernike_freq_interp'])
            
        dish_diam = beam.dish_diam
        lmbd = c/freq
        eta = beam_interp.ETA[0,0].values #Assume ETA is the same for all pol and coef. Only a function of freq. Need to update zpc.zarr format.
        uv_cell_size = 1/(beam_parms['cell_size']*beam_parms['image_size'])
        zernike_cell = (2.0*uv_cell_size*lmbd)/(dish_diam*eta)
        
        
        if (pa != pa_prev) or (freq != freq_prev) :
            beam_parms['parallactic_angle'] = pa
            image_size = (np.ceil(np.abs(2.0/zernike_cell))).astype(int)
            x_grid, y_grid = _compute_rot_coords(image_size,zernike_cell,pa)
            
            r_grid = np.sqrt(x_grid**2 + y_grid**2)
            
            zernike_size = np.array(x_grid.shape)
        
            ic_z = zernike_size//2
            include_last = (zernike_size%2).astype(int)
        
        #assert zernike_size[0] < beam_parms['conv_size'][0] and zernike_size[1] < gcf_parms['conv_size'][1], "The convolution size " + str(gcf_parms['conv_size']) +" is smaller than the aperture image " + zernike_size + " . Increase conv_size"
        
        #print('x_grid',x_grid)
        #print('y_grid',y_grid)
        
        
        start = time.time()
        for i_pol,pol in enumerate(beam_parms['needed_pol']):
            a = _generate_zernike_surface(beam_interp.ZPC[pol,:].values,x_grid,y_grid)
            a[r_grid > 1] = 0
            j_planes[i_pa, i_chan,i_pol,ic[0]-ic_z[0]:ic[0]+ic_z[0]+include_last[0],ic[1]-ic_z[1]:ic[1]+ic_z[1]+include_last[1]] = a
            j_planes[i_pa, i_chan, i_pol,:,:] = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(j_planes[i_pa, i_chan,i_pol,:,:])))/(beam_parms['image_size'][0]*beam_parms['image_size'][1])
        #print('One pol set',time.time()-start)
        
        #Normalize Jones
        if 3 not in beam_parms['needed_pol']:
            #P_max = np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==0),j_planes_shape[3]//2,j_planes_shape[4]//2])
            P_max = np.max(np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==0),:,:]))
            Q_max = P_max
        elif 0 not in beam_parms['needed_pol']:
            #Q_max = np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==3),j_planes_shape[3]//2,j_planes_shape[4]//2])
            Q_max = np.max(np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==3),:,:]))
            P_max = Q_max
        else:
            #P_max = np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==0),j_planes_shape[3]//2,j_planes_shape[4]//2])
            P_max = np.max(np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==0),:,:]))
            #Q_max = np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==3),j_planes_shape[3]//2,j_planes_shape[4]//2])
            Q_max = np.max(np.abs(j_planes[i_pa, i_chan, np.where(beam_parms['needed_pol']==3),:,:]))
        j_planes[i_pa, i_chan,:,:,:] = j_planes[i_pa, i_chan,:,:,:]*2/(P_max+Q_max)
        
        pa_prev = pa
        freq_prev = freq
    return j_planes#np.zeros((1,4,2048,2048),dtype=np.complex128)







#####################################################################################

##################Currently Not used functions##################
#@jit(nopython=True,cache=True)
def _outer_product(B1,B2,norm,conj):
    '''
    Input
    B1 2 x 2 x m x n array
    B2 2 x 2 x m x n array
    Output
    M 4 x 4 x m x n
    '''
    
    #assert B1.shape==B2.shape
    
    s = B1.shape
    
    M = np.zeros((4,4,s[2],s[3]),dtype=np.complex128)
    
    indx_b1 = np.array([[[0,0],[0,0],[0,1],[0,1]],[[0,0],[0,0],[0,1],[0,1]],[[1,0],[1,0],[1,1],[1,1]],[[1,0],[1,0],[1,1],[1,1]]])
    indx_b2 = np.array([[[0,0],[0,1],[0,0],[0,1]],[[1,0],[1,1],[1,0],[1,1]],[[0,0],[0,1],[0,0],[0,1]],[[1,0],[1,1],[1,0],[1,1]]])
    #print(indx_b1.shape)
    
    
    for i in range(4):
        for j in range(4):
            #print(indx_b1[i,j,:], ',*,', indx_b2[i,j,:])
            if conj:
                M[i,j,:,:] = B1[indx_b1[i,j,0],indx_b1[i,j,1],:,:] * B2[indx_b2[i,j,0],indx_b2[i,j,1],:,:].conj().T
            else:
                M[i,j,:,:] = B1[indx_b1[i,j,0],indx_b1[i,j,1],:,:] * B2[indx_b2[i,j,0],indx_b2[i,j,1],:,:]
                
            if norm:
                M[i,j,:,:] = M[i,j,:,:]/np.max(np.abs(M[i,j,:,:]))
    
    #print(M.shape)
    return(M)


def _outer_product_conv(B1,B2):
    
#    Input
#    B1 2 x 2 x m x n array
#    B2 2 x 2 x m x n array
#    Output
#    M 4 x 4 x m x n
    
    #assert B1.shape==B2.shape
    
    s = B1.shape
    
    M = np.zeros((4,4,s[2],s[3]),dtype=np.complex128)
    
    indx_b1 = np.array([[[0,0],[0,0],[0,1],[0,1]],[[0,0],[0,0],[0,1],[0,1]],[[1,0],[1,0],[1,1],[1,1]],[[1,0],[1,0],[1,1],[1,1]]])
    indx_b2 = np.array([[[0,0],[0,1],[0,0],[0,1]],[[1,0],[1,1],[1,0],[1,1]],[[0,0],[0,1],[0,0],[0,1]],[[1,0],[1,1],[1,0],[1,1]]])
    
    for i in range(4):
        for j in range(4):
            M[i,j,:,:] = signal.fftconvolve(B1[indx_b1[i,j,0],indx_b1[i,j,1],:,:], B2[indx_b2[i,j,0],indx_b2[i,j,1],:,:],mode='same')
    
    #print(M.shape)
    return(M)

    
def _make_flat(B):
    '''
    B 2x2xmxn
    B_flat 2mx2n
    '''
    s = B.shape
    B_flat = np.zeros((s[2]*s[0],s[3]*s[1]),dtype=complex)
    
    
    for i in range(s[0]):
        for j in range(s[1]):
            i_start = i*s[2]
            i_end = (i+1)*s[3]
            j_start = j*s[2]
            j_end = (j+1)*s[3]
            B_flat[i_start:i_end,j_start:j_end] = B[i,j,:,:]
            #print(B[i,j,1024,1024],np.abs(B[i,j,1024,1024]))
    return B_flat
    
    
def _make_flat_casa(B):
    '''
    B mxnx16
    B_flat 4mx4n
    '''
    s = B.shape
    B_flat = np.zeros((s[0]*4,s[1]*4),dtype=complex)
    
    #indx = np.array([[0,0],[1,0],[2,0],[3,0],[0,1],[1,1],[2,1],[3,1],[0,2],[1,2],[2,2],[3,2],[0,3],[1,3],[2,3],[3,3]])
    indx = np.array([[0,0],[0,1],[0,2],[0,3],[1,0],[1,1],[1,2],[1,3],[2,0],[2,1],[2,2],[2,3],[3,0],[3,1],[3,2],[3,3]]) #saved as rows
    
    for c,i in enumerate(indx):
        #print(c,i)
        i_start = i[0]*s[0]
        i_end = (i[0]+1)*s[0]
        j_start = i[1]*s[1]
        j_end = (i[1]+1)*s[1]
        B_flat[i_start:i_end,j_start:j_end] = B[:,:,c].T
        #print(B[1024,1024,c],np.abs(B[1024,1024,c]))
    return B_flat



def _make_flat_casa(B):
    '''
    B mxnx16
    B_flat 4mx4n
    '''
    s = B.shape
    B_flat = np.zeros((s[0]*4,s[1]*4),dtype=complex)
    
    #indx = np.array([[0,0],[1,0],[2,0],[3,0],[0,1],[1,1],[2,1],[3,1],[0,2],[1,2],[2,2],[3,2],[0,3],[1,3],[2,3],[3,3]])
    indx = np.array([[0,0],[0,1],[0,2],[0,3],[1,0],[1,1],[1,2],[1,3],[2,0],[2,1],[2,2],[2,3],[3,0],[3,1],[3,2],[3,3]]) #saved as rows
    
    for c,i in enumerate(indx):
        #print(c,i)
        i_start = i[0]*s[0]
        i_end = (i[0]+1)*s[0]
        j_start = i[1]*s[1]
        j_end = (i[1]+1)*s[1]
        B_flat[i_start:i_end,j_start:j_end] = B[:,:,c].T
        #print(B[1024,1024,c],np.abs(B[1024,1024,c]))
    return B_flat



 #   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


sirius_logger_name = 'sirius'
import sys
import logging
from datetime import datetime
formatter = logging.Formatter("[%(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s")
from dask.distributed import get_worker
from dask.distributed import WorkerPlugin
import dask

def _get_sirius_logger(name = sirius_logger_name):
    '''
    Will fisrt try to get worker logger. If this fails graph construction logger is returned.
    '''
    from dask.distributed import get_worker
    try:
        worker = get_worker()
    except:
        return logging.getLogger(name)
    
    try:
        logger = worker.plugins['sirius_worker_logger'].get_logger()
        return logger
    except:
        return logging.getLogger()

def _setup_sirius_logger(log_to_term=False,log_to_file=True,log_file='sirius_', level='INFO', name=sirius_logger_name):
    """To setup as many loggers as you want"""
    logger = logging.getLogger(name)
    logger.setLevel(logging.getLevelName(level))
    
    if log_to_term:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    
    if log_to_file:
        log_file = log_file+datetime.today().strftime('%Y%m%d_%H%M%S')+'.log'
        handler = logging.FileHandler(log_file)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        
    return logger
    
def _get_sirius_worker_logger_name(name=sirius_logger_name):
    worker_log_name = name + '_' + str(get_worker().id)
    return worker_log_name

class _sirius_worker_logger_plugin(WorkerPlugin):
    def __init__(self,log_parms):
        self.log_to_term=log_parms['log_to_term']
        self.log_to_file=log_parms['log_to_file']
        self.log_file=log_parms['log_file']
        self.level=log_parms['log_level']
        self.logger = None
        print(self.log_to_term,self.log_to_file,self.log_file,self.level)
        
    def get_logger(self):
        return self.logger
        
    def setup(self, worker: dask.distributed.Worker):
        "Run when the plugin is attached to a worker. This happens when the plugin is registered and attached to existing workers, or when a worker is created after the plugin has been registered."
        self.logger = _setup_sirius_worker_logger(self.log_to_term,self.log_to_file,self.log_file,self.level)

def _setup_sirius_worker_logger(log_to_term,log_to_file,log_file, level):
    parallel_logger_name = _get_sirius_worker_logger_name()
    
    logger = logging.getLogger(parallel_logger_name)
    logger.setLevel(logging.getLevelName(level))
    
    if log_to_term:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    
    if log_to_file:
        log_file = log_file + '_' + str(get_worker().id) + '_' + datetime.today().strftime('%Y%m%d_%H%M%S') + '.log'
        handler = logging.FileHandler(log_file)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger




#http://www.aoc.nrao.edu/evla/geninfo/memoseries/evlamemo113.pdf
#https://library.nrao.edu/public/memos/evla/EVLAM_195.pdf
#https://open-bitbucket.nrao.edu/projects/CASA/repos/casa6/browse/casatools/src/code/synthesis/TransformMachines/PBMath1DEVLA.cc
'''
    //Float mag = 1.21;  // half-squint magnitude in arcmin at 1 GHz)
    feedConf_p["L"] = (-185.9)*C::pi/180.0;    // squint orientation, rads, North of +AZ axis
    feedConf_p["S"] = (-11.61)*C::pi/180.0;
    feedConf_p["C"]=(-104.8)*C::pi/180.0;
    feedConf_p["X"]=(-113.7)*C::pi/180.0;
    feedConf_p["U"]=(42.4)*C::pi/180.0;
    feedConf_p["K"]=(64.4)*C::pi/180.0;
    feedConf_p["A"]=(106.9)*C::pi/180.0;
    feedConf_p["Q"]=(85.5)*C::pi/180.0;
'''
def txt_to_zarr(filename,freq_to_hertz,dish_diam,max_coef):
    #from cngi.dio import write_zarr
    import xarray as xr
    import numpy as np
    import dask.array as da
    import csv
    from datetime import date
    import matplotlib.pyplot as plt
    
    band_mode = False
    coef_mode = False
    band = None
    
    band_name = []
    band_name_axis = []
    band_start = []
    band_end = []
    
    pol_coef = []
    
    freq = []
    
    
    txt_list = []
    with open(filename, newline='') as txtfile:
        lines = txtfile.readlines()
        for i,line in enumerate(lines):
            #print(i,line)
            if "# Band definitions" in line:
                band_mode = True
                coef_mode = False
            if "# Coefficients" in line:
                coef_mode = True
                band_mode = False
            
            if band_mode and "# Band definitions" not in line:
                split_line = line.split(' ')
                band_name.append(split_line[0])
                band_start.append(float(split_line[1]))
                band_end.append(float(split_line[2][:-1]))
             
            if coef_mode and "# Coefficients" not in line:
                if line[0] in band_name:
                    band = line[0]
                
                if line[0] not in band_name:
                    band_name_axis.append(band)
                    start_indx = line.index('{')+1
                    end_indx =  line.index('}')
                    split_line_coef = line[start_indx:end_indx].split(',')
                    start_indx = line.index('[')+1
                    end_indx =  line.index(']')
                    split_line_freq = line[start_indx:end_indx]
                    #print(len(split_line))
                    
                    coef_temp = np.zeros((max_coef,))
                    freq.append(float(split_line_freq)*10**6)
                    
                    assert len(split_line_coef) <= max_coef, 'Number of coefficients exceed max_coef'

                    for j,coef in enumerate(split_line_coef):
                        coef_temp[j] = float(coef)
                        
                    pol_coef.append(coef_temp)
    
    pol_coef = np.array(pol_coef)
    freq = np.array(freq)
    band = np.array(band_name_axis)
    #print(pol_coef.shape)
    #print(freq.shape)
                    
    pc_dataset = xr.Dataset()
    coords = {'pol':[5],'chan':freq,'coef_indx':np.arange(max_coef),'band':('chan',band)}
    pc_dataset = pc_dataset.assign_coords(coords)
    pc_dataset['BPC'] = xr.DataArray(pol_coef[:,None,:], dims=['chan','pol','coef_indx'])
    pc_dataset['ETA'] = xr.DataArray(np.zeros(pc_dataset['BPC'].shape), dims=['chan','pol','coef_indx'])
    pc_dataset.attrs['pc_file_name'] = filename.partition('/')[2]
    pc_dataset.attrs['telescope_name'] = filename.partition('/')[2].partition('_')[0]
    pc_dataset.attrs['conversion_date'] = str(date.today())
    pc_dataset.attrs['dish_diam'] = dish_diam
    
    if pc_dataset.attrs['telescope_name'].lower() == 'evla':
        pc_dataset.attrs['max_rad_1GHz'] = 0.8564*np.pi/180
    elif pc_dataset.attrs['telescope_name'].lower() == 'ngvla':
        pc_dataset.attrs['max_rad_1GHz'] = 1.5*np.pi/180
    elif pc_dataset.attrs['telescope_name'].lower() == 'alma':
        pc_dataset.attrs['max_rad_1GHz'] = 1.784*np.pi/180
    elif pc_dataset.attrs['telescope_name'].lower() == 'aca':
        pc_dataset.attrs['max_rad_1GHz'] = 3.568*np.pi/180
        
    #write_zarr(pc_dataset,filename.split('.')[0]+'.bpc.zarr')
    xr.Dataset.to_zarr(pc_dataset,filename.split('.')[0]+'.bpc.zarr',mode='w')
    
    print(pc_dataset)
    
if __name__ == '__main__':
    import shutil
    #Remove all . in name except for last (before .csv)
    filenames = ['EVLA_.txt']
    dish_diams = [25]
    freq_to_hertz = 10**6
    max_coefs =[5]
    for filename,dish_diam,max_coef in zip(filenames,dish_diams,max_coefs):
        print(filename)
        txt_to_zarr('data/'+filename,freq_to_hertz,dish_diam,max_coef)
        
        try:
            shutil.make_archive('data/'+filename[:-4]+'.bpc.zarr', 'zip', 'data/'+filename[:-4]+'.bpc.zarr')
        except:
            print('Cant compress','data/'+filename[:-4]+'.bpc.zarr')



#   Copyright 2019 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import xarray as xr
import dask.array as da
from astropy.timeseries import TimeSeries
from astropy.time import Time
from astropy import units as u
from collections import Counter
import time
import dask
import os
import shutil
import logging

def make_time_xda(
    time_start="2019-10-03T19:00:00.000", time_delta=3600, n_samples=10, n_chunks=4
):
    """
    Create a time series xarray array.
    Parameters
    ----------
    -------
    time_xda : xarray.DataArray
    """
    ts = np.array(
        TimeSeries(
            time_start=time_start, time_delta=time_delta * u.s, n_samples=n_samples
        ).time.value
    )
    chunksize = int(np.ceil(n_samples / n_chunks))
    time_da = da.from_array(ts, chunks=chunksize)
    #print("Number of chunks ", len(time_da.chunks[0]))

    time_xda = xr.DataArray(
        data=time_da, dims=["time"], attrs={"time_delta": float(time_delta)}
    )
    
    time_xda.attrs['time_delta'] = time_delta

    return time_xda


def make_chan_xda(
    spw_name="sband",
    freq_start=3 * 10**9,
    freq_delta=0.4 * 10**9,
    freq_resolution=0.01 * 10**9,
    n_channels=3,
    n_chunks=3,
):
    """
    Create a channel frequencies xarray array.
    Parameters
    ----------
    -------
    chan_xda : xarray.DataArray
    """
    freq_chan = (np.arange(0, n_channels) * freq_delta + freq_start).astype(
        float
    )  # astype(float) needed for interfacing with CASA simulator.
    chunksize = int(np.ceil(n_channels / n_chunks))
    chan_da = da.from_array(freq_chan, chunks=chunksize)
    #print("Number of chunks ", len(chan_da.chunks[0]))

    chan_xda = xr.DataArray(
        data=chan_da,
        dims=["chan"],
        attrs={
            "freq_resolution": float(freq_resolution),
            "spw_name": spw_name,
            "freq_delta": float(freq_delta),
        },
    )
    return chan_xda


from dask.distributed import get_client


def create_mxds(
    xds0,
    time_xda,
    chan_xda,
    pol,
    tel_xds,
    phase_center_names,
    phase_center_ra_dec,
    auto_corr,
    save_parms,
):
    n_row, n_chan, n_pol = xds0['DATA'].shape
    n_time = len(time_xda)
    n_baseline = int(n_row/n_time)
    
    
    assert (n_row/n_time - int(n_row/n_time)) == 0, 'n_time*n_basline does not equal n_row'

    #Create mxds
    mxds = xr.Dataset()
    coords = {'polarization_ids':np.array([0]), 'spw_ids': np.array([0])}
    mxds = mxds.assign_coords(coords)


    ############## Create main table xds ##############
    coords = {'chan':chan_xda.values, 'pol': pol, 'pol_id':np.array([0]), 'spw_id': np.array([0])}
    
    chunks = {"row": (xds0['DATA'].chunks[0][0],), "chan": (xds0['DATA'].chunks[1][0],), "pol": (xds0['DATA'].chunks[2][0],), "uvw": (3,)}
    chunks =  (xds0['DATA'].chunks[0][0], xds0['DATA'].chunks[1][0],xds0['DATA'].chunks[2][0])


    empty_data_column = da.zeros(xds0['DATA'].shape, chunks=chunks, dtype="complex")
    flag_rows = da.zeros(n_row,chunks=chunks[0], dtype=np.bool)
    flags = da.zeros(xds0['DATA'].shape,chunks=chunks, dtype=np.bool) # don't flag any of the data yet
    zero_column = da.zeros(n_row, chunks=chunks[0], dtype="int32")
    exposures = da.full((n_row,), time_xda.time_delta, chunks=chunks[0], dtype="float64") # fill with input in units of the input array, which we

    xds0 = xds0.assign_coords(coords)
    xds0['CORRECTED_DATA'] = xds0['DATA']
    xds0['MODEL_DATA'] = xr.DataArray(empty_data_column, dims=['row','chan','pol'])
    xds0['FLAG'] = xr.DataArray(flags, dims=['row','chan','pol'])
    xds0['FLAG_ROW'] = xr.DataArray(flag_rows, dims=['row'])
    xds0['DATA_DESC_ID'] = xr.DataArray(zero_column, dims=['row'])
    xds0['ARRAY_ID'] = xr.DataArray(zero_column, dims=['row'])
    xds0['FEED1'] = xr.DataArray(zero_column, dims=['row'])
    xds0['FEED2'] = xr.DataArray(zero_column, dims=['row'])
    xds0['OBSERVATION_ID'] = xr.DataArray(zero_column, dims=['row'])
    xds0['PROCESSOR_ID'] = xr.DataArray(zero_column, dims=['row'])
    xds0['SCAN_NUMBER'] = xr.DataArray(zero_column, dims=['row'])
    xds0['STATE_ID'] = xr.DataArray(zero_column, dims=['row'])
    xds0['EXPOSURE'] = xr.DataArray(exposures, dims=['row'])
    xds0['INTERVAL'] = xr.DataArray(exposures, dims=['row'])
    xds0['TIME_CENTROID'] = xds0['TIME']
    
    xds0.attrs['MS_VERSION'] = 2.0
    from sirius_data._ms_column_descriptions_dicts import main_column_description
    xds0.attrs['column_descriptions'] = main_column_description
    xds0.attrs['info'] = {'type': 'Measurement Set', 'subType': 'simulator', 'readme': 'This is a MeasurementSet Table holding measurements from a Telescope\nThis is a MeasurementSet Table holding simulated astronomical observations\n'}
    xds0.attrs['bad_cols'] = ['FLAG_CATEGORY']
    mxds.attrs['xds0'] = xds0
    
    from sirius._sirius_utils._ms_utils import _sub_table_xds
    mxds = _sub_table_xds(mxds,time_xda,chan_xda,pol,tel_xds,phase_center_names,phase_center_ra_dec,auto_corr)
    return mxds
    
def write_to_ms_daskms_and_sim_tool(
    vis_xds,
    time_xda,
    chan_xda,
    pol,
    tel_xds,
    phase_center_names,
    phase_center_ra_dec,
    auto_corr,
    save_parms,
):
    """
    Write out a MeasurementSet to disk using dask-ms

    This first implementation is kept only temporarily, until performance comparisons are completed.
    """

    start = time.time()
    from casatools import simulator
    from casatasks import mstransform

    n_row, n_chan, n_pol = vis_xds.DATA.shape
    n_time = len(time_xda.data)
    
    print(n_row, n_chan, n_pol)
    print(vis_xds)

    sm = simulator()

    ant_pos = tel_xds.ANT_POS.values
    os.system("rm -rf " + save_parms["ms_name"])
    sm.open(ms=save_parms["ms_name"])

    ###########################################################################################################################
    ## Set the antenna configuration
    sm.setconfig(
        telescopename=tel_xds.telescope_name,
        x=ant_pos[:, 0],
        y=ant_pos[:, 1],
        z=ant_pos[:, 2],
        dishdiameter=tel_xds.DISH_DIAMETER.values,
        mount=["alt-az"],
        antname=list(
            tel_xds.ant_name.values
        ),  # CASA can't handle an array of antenna names.
        coordsystem="global",
        referencelocation=tel_xds.site_pos[0],
    )

    ## Set the polarization mode (this goes to the FEED subtable)
    from sirius_data._constants import pol_codes_RL, pol_codes_XY, pol_str
    from sirius._sirius_utils._array_utils import _is_subset

    if _is_subset(pol_codes_RL, pol):  # ['RR','RL','LR','LL']
        sm.setfeed(mode="perfect R L", pol=[""])
    elif _is_subset(pol_codes_XY, pol):  # ['XX','XY','YX','YY']
        sm.setfeed(mode="perfect X Y", pol=[""])
    else:
        assert False, print(
            "Pol selection invalid, must either be subset of [5,6,7,8] or [9,10,11,12] but is ",
            pol,
        )

    sm.setspwindow(
        spwname=chan_xda.spw_name,
        freq=chan_xda.data[0].compute(),
        deltafreq=chan_xda.freq_delta,
        freqresolution=chan_xda.freq_resolution,
        nchannels=len(chan_xda),
        refcode="LSRK",
        stokes=" ".join(pol_str[pol]),
    )

    if auto_corr:
        sm.setauto(autocorrwt=1.0)
    else:
        sm.setauto(autocorrwt=0.0)

    mjd = Time(time_xda.data[0].compute(), scale="utc")
     
    integration_time = time_xda.attrs['time_delta']* u.second
    
    print('mjd',mjd)

    start_time = (mjd - (integration_time / 2 + 37 * u.second)).mjd
    start_time_dict = {
        "m0": {"unit": "d", "value": start_time},
        "refer": "UTC",
        "type": "epoch",
    }

    sm.settimes(
        integrationtime=integration_time.value,
        usehourangle=False,
        referencetime=start_time_dict,
    )

    fields_set = []
    field_time_count = Counter(phase_center_names)

    # print(field_time_count,phase_center_names)
    if len(phase_center_names) == 1:  # Single field case
        field_time_count[list(field_time_count.keys())[0]] = n_time

    start_time = 0
    for i, ra_dec in enumerate(
        phase_center_ra_dec
    ):  # In future make phase_center_ra_dec a unique list
        if phase_center_names[i] not in fields_set:
            dir_dict = {
                "m0": {"unit": "rad", "value": ra_dec[0]},
                "m1": {"unit": "rad", "value": ra_dec[1]},
                "refer": "J2000",
                "type": "direction",
            }
            sm.setfield(sourcename=phase_center_names[i], sourcedirection=dir_dict)
            fields_set.append(phase_center_names[i])

            stop_time = (
                start_time
                + integration_time.value * field_time_count[phase_center_names[i]]
            )
            sm.observe(
                sourcename=phase_center_names[i],
                spwname=chan_xda.spw_name,
                starttime=str(start_time) + "s",
                stoptime=str(stop_time) + "s",
            )
            start_time = stop_time

    print("Meta data creation ", time.time() - start)

    # print(vis_data.shape)
    # print(n_row,n_time, n_baseline, n_chan, n_pol)

    start = time.time()
    # This code will most probably be moved into simulation if we get rid of row time baseline split.
    #vis_data_reshaped = vis_xds.DATA.data.reshape((n_row, n_chan, n_pol))
    #uvw_reshaped = vis_xds.UVW.data.reshape((n_row, 3))
    #weight_reshaped = vis_xds.WEIGHT.data.reshape((n_row, n_pol))
    #sigma_reshaped = vis_xds.SIGMA.data.reshape((n_row, n_pol))

    #print("reshape time ", time.time() - start)
    # weight_spectrum_reshaped = np.tile(weight_reshaped[:,None,:],(1,n_chan,1))

    #    print(weight_reshaped.compute().shape)
    #    print(sigma_reshaped.compute().shape)
    #    print(weight_reshaped)
    #    print(sigma_reshaped)

    # dask_ddid = da.full(n_row, 0, chunks=chunks['row'], dtype=np.int32)

    # print('vis_data_reshaped',vis_data_reshaped)

    start = time.time()
    from daskms import xds_to_table, xds_from_ms, Dataset

    # print('vis_data_reshaped.chunks',vis_data_reshaped.chunks)
    row_id = da.arange(n_row, chunks=vis_xds.DATA.chunks[0], dtype="int32")

    dataset = Dataset(
        {
            "DATA": (("row", "chan", "corr"), vis_xds.DATA.data),
            "CORRECTED_DATA": (("row", "chan", "corr"), vis_xds.DATA.data),
            "UVW": (("row", "uvw"), vis_xds.UVW.data),
            "SIGMA": (("row", "pol"), vis_xds.SIGMA.data),
            "WEIGHT": (("row", "pol"), vis_xds.WEIGHT.data),
            "ROWID": (("row",), row_id),
        }
    )
    # ,'WEIGHT_SPECTRUM': (("row","chan","pol"), weight_spectrum_reshaped)
    ms_writes = xds_to_table(dataset, save_parms["ms_name"], columns="ALL")


    if save_parms["DAG_name_write"]:
        dask.visualize(ms_writes, filename=save_parms["DAG_name_write"])

    start = time.time()
    dask.compute(ms_writes)
    print("*** Dask compute time", time.time() - start)

    sm.close()

    from casatasks import flagdata

    flagdata(vis=save_parms["ms_name"], mode="unflag")

    
        
def read_zarr(
    infile,
    sel_xds=None,
    chunks=None,
    consolidated=True,
    overwrite_encoded_chunks=True,
    **kwargs,
):
    """
    Read zarr format Visibility data from disk to xarray Dataset

    Parameters
    ----------
    infile : str
        input Visibility filename
    sel_xds : string or list
        Select the ddi to open, for example ['xds0','xds1'] will open the first two ddi. Default None returns everything
    chunks : dict
        sets specified chunk size per dimension. Dict is in the form of 'dim':chunk_size, for example {'time':100, 'baseline':400, 'chan':32, 'pol':1}.
        Default None uses the original zarr chunking.
    consolidated : bool
        use zarr consolidated metadata capability. Only works for stores that have already been consolidated. Default True works with datasets
        produced by convert_ms which automatically consolidates metadata.
    overwrite_encoded_chunks : bool
        drop the zarr chunks encoded for each variable when a dataset is loaded with specified chunk sizes.  Default True, only applies when chunks
        is not None.
        
    Returns
    -------
    xarray.core.dataset.Dataset
        New xarray Dataset of Visibility data contents
    """
    import os
    import numpy as np
    import cngi._utils._io as xdsio
    from xarray import open_zarr

    if chunks is None:
        chunks = "auto"
        #overwrite_encoded_chunks = False
    #print('overwrite_encoded_chunks',overwrite_encoded_chunks)

    infile = os.path.expanduser(infile)
    if sel_xds is None:
        sel_xds = os.listdir(infile)
    sel_xds = list(np.atleast_1d(sel_xds))
    
    
    #print(os.path.join(infile, 'DDI_INDEX'))
    mxds = open_zarr(os.path.join(infile, 'DDI_INDEX'), chunks=chunks,consolidated=consolidated,overwrite_encoded_chunks=overwrite_encoded_chunks)

    for part in os.listdir(os.path.join(infile, "global")):
        xds_temp = open_zarr(os.path.join(infile, 'global/'+part), chunks=chunks,
                                                                     consolidated=consolidated,
                                                                     overwrite_encoded_chunks=overwrite_encoded_chunks)
        xds_temp = _fix_dict_for_ms(part,xds_temp)
        mxds.attrs[part] = xds_temp.compute()

    for part in os.listdir(infile):
        if ('xds' in part) and (part in sel_xds):
            xds_temp = open_zarr(os.path.join(infile, part), chunks=chunks,
                                                                     consolidated=consolidated,
                                                                     overwrite_encoded_chunks=overwrite_encoded_chunks)
            xds_temp = _fix_dict_for_ms(part,xds_temp)
            mxds.attrs[part] = xds_temp
            
    return mxds
 
    
def write_zarr(mxds, outfile, chunks_on_disk=None, partition=None, consolidated=True, compressor=None, overwrite=True, graph_name='write_zarr'):
    """
    Write xarray dataset to zarr format on disk. When chunks_on_disk is not specified the chunking in the input dataset is used.
    When chunks_on_disk is specified that dataset is saved using that chunking.

    Parameters
    ----------
    mxds : xarray.core.dataset.Dataset
        Dataset of dataset to write to disk
    outfile : str
        outfile filename, generally ends in .zarr
    chunks_on_disk : dict of int
        A dictionary with the chunk size that will be used when writing to disk. For example {'time': 20, 'chan': 6}.
        If chunks_on_disk is not specified the chunking of dataset will be used.
    partition : str or list
        Name of partition xds to write into outfile (from the mxds attributes section). Overwrites existing partition of same name.
        Default None writes entire mxds
    compressor : numcodecs.blosc.Blosc
        The blosc compressor to use when saving the converted data to disk using zarr.
        If None the zstd compression algorithm used with compression level 2.
    graph_name : string
        The time taken to execute the graph and save the dataset is measured and saved as an attribute in the zarr file.
        The graph_name is the label for this timing information.

    Returns
    -------
    """
    import xarray as xr
    import zarr
    import time
    from numcodecs import Blosc
    from itertools import cycle
    import os
    import numpy as np

    if compressor is None:
        compressor = Blosc(cname='zstd', clevel=2, shuffle=0)

    if partition is None:
        partition = list(mxds.attrs.keys())
    partition = list(np.atleast_1d(partition))
        
    if overwrite:
        try:
            os.remove(outfile)
        except IsADirectoryError:
            shutil.rmtree(outfile)
        except FileNotFoundError:
            pass
        os.system('mkdir ' + outfile)
    else:
        assert not os.path.isfile(outfile), 'vis.zarr folder already exists. Set overwrite to True.'
        
    ddi_indx_xds = xr.Dataset()
    ddi_indx_xds['polarization_ids'] = mxds['polarization_ids']
    ddi_indx_xds['spw_ids'] = mxds['spw_ids']
    encoding = dict(zip(list(ddi_indx_xds.data_vars), cycle([{'compressor': compressor}])))
    xr.Dataset.to_zarr(ddi_indx_xds, store=outfile+'/DDI_INDEX', mode='w', encoding=encoding,consolidated=consolidated)
        
    for xds_name in partition:
        if "xds" in xds_name:
            xds_outfile = outfile + '/' + xds_name
            xds_for_disk = mxds.attrs[xds_name]
            if chunks_on_disk is not None:
                xds_for_disk = xds_for_disk.chunk(chunks=chunks_on_disk)
        else:
            xds_outfile = outfile + '/global/' + xds_name
            xds_for_disk = mxds.attrs[xds_name]
            
        xds_for_disk = _fix_dict_for_zarr(xds_name, xds_for_disk)
            
        # Create compression encoding for each datavariable
        encoding = dict(zip(list(xds_for_disk.data_vars), cycle([{'compressor': compressor}])))
        start = time.time()

        # Consolidated is set to False so that the timing information is included in the consolidate metadata.
        xr.Dataset.to_zarr(xds_for_disk, store=xds_outfile, mode='w', encoding=encoding,consolidated=False)
        time_to_calc_and_store = time.time() - start
        print('Time to store and execute graph for ', xds_name, graph_name, time_to_calc_and_store)
      
        #Add timing information
        #dataset_group = zarr.open_group(xds_outfile, mode='a')
        #dataset_group.attrs[graph_name+'_time'] = time_to_calc_and_store
            
        if consolidated == True:
            zarr.consolidate_metadata(xds_outfile)


def _fix_dict_for_ms(name, xds):
    xds.attrs['column_descriptions'] = xds.attrs['column_descriptions'][0]
    xds.attrs['info'] = xds.attrs['info'][0]

    if "xds" in name:
        xds.column_descriptions['UVW']['shape'] = np.array(xds.column_descriptions['UVW']['shape'].split(',')).astype(int)

    if "SPECTRAL_WINDOW" == name:
        #print('2.',xds.column_descriptions)
        xds.column_descriptions['CHAN_FREQ']['keywords']['MEASINFO']['TabRefCodes'] = np.array(xds.column_descriptions['CHAN_FREQ']['keywords']['MEASINFO']['TabRefCodes'].split(',')).astype(int)
        xds.column_descriptions['REF_FREQUENCY']['keywords']['MEASINFO']['TabRefCodes'] =  np.array(xds.column_descriptions['REF_FREQUENCY']['keywords']['MEASINFO']['TabRefCodes'].split(',')).astype(int)
        
    if "ANTENNA" == name:
        xds.column_descriptions['OFFSET']['shape'] = np.array(xds.column_descriptions['OFFSET']['shape'].split(',')).astype(int)
        xds.column_descriptions['POSITION']['shape'] = np.array(xds.column_descriptions['POSITION']['shape'].split(',')).astype(int)
    
    if "FEED" == name:
        xds.column_descriptions['POSITION']['shape'] = np.array(xds.column_descriptions['POSITION']['shape'].split(',')).astype(int)

    if "OBSERVATION" == name:
        xds.column_descriptions['TIME_RANGE']['shape'] = np.array(xds.column_descriptions['TIME_RANGE']['shape'].split(',')).astype(int)

    return xds
    
def _fix_dict_for_zarr(name, xds):
    xds.attrs['column_descriptions'] = [xds.attrs['column_descriptions']]
    xds.attrs['info'] = [xds.attrs['info']]
    
    if "xds" in name:
        xds.column_descriptions[0]['UVW']['shape'] = ','.join(map(str, xds.column_descriptions[0]['UVW']['shape']))

    if "SPECTRAL_WINDOW" == name:
        #print('1.',xds.column_descriptions)
        xds.column_descriptions[0]['CHAN_FREQ']['keywords']['MEASINFO']['TabRefCodes'] = ','.join(map(str, xds.column_descriptions[0]['CHAN_FREQ']['keywords']['MEASINFO']['TabRefCodes']))
        xds.column_descriptions[0]['REF_FREQUENCY']['keywords']['MEASINFO']['TabRefCodes'] = ','.join(map(str, xds.column_descriptions[0]['REF_FREQUENCY']['keywords']['MEASINFO']['TabRefCodes']))
    
    if "ANTENNA" == name:
        xds.column_descriptions[0]['OFFSET']['shape'] = ','.join(map(str, xds.column_descriptions[0]['OFFSET']['shape']))
        xds.column_descriptions[0]['POSITION']['shape'] = ','.join(map(str, xds.column_descriptions[0]['POSITION']['shape']))
    
    if "FEED" == name:
        xds.column_descriptions[0]['POSITION']['shape'] = ','.join(map(str, xds.column_descriptions[0]['POSITION']['shape']))

    if "OBSERVATION" == name:
        xds.column_descriptions[0]['TIME_RANGE']['shape'] = ','.join(map(str, xds.column_descriptions[0]['TIME_RANGE']['shape']))
        
    return xds












################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################
################################################################################################################################################

    '''
    min_memory_for_a_thread, _, _ = _get_schedular_info()
    #chunksize_row = _calc_optimal_ms_chunk_shape(min_memory_for_a_thread,(n_row,),64,'TIME')
    chunksize_row = _calc_optimal_ms_chunk_shape(min_memory_for_a_thread,(n_row,),128,'MODEL_DATA')
    empty_data_column = da.zeros(xds0['DATA'].shape, chunks=(chunksize_row,-1,-1), dtype="complex")
    
    chunksize_row = _calc_optimal_ms_chunk_shape(min_memory_for_a_thread,(n_row,),8,'FLAG_ROW')
    flag_rows = da.zeros(n_row,chunks=chunksize_row, dtype=np.bool)

    chunksize_row = _calc_optimal_ms_chunk_shape(min_memory_for_a_thread,xds0['DATA'].shape,8,'FLAG')
    flags = da.zeros(xds0['DATA'].shape,chunks=(chunksize_row,-1,-1), dtype=np.bool) # don't flag any of the data yet

   
#    ddid = da.zeros(n_row, chunks=chunksize_row, dtype="int32") # we run this function on only a single DDI at a time
#    array_ids = da.zeros_like(ddid) # currently don't support subarrays, so only one array ID assigned
#    feeds = da.zeros_like(ddid)  # not supporting different feed types
#    observation_ids = da.zeros_like(ddid) # this function is also only run for a single observation at once
#    processor_ids = da.zeros_like(ddid) # currently don't support group processing
#    scan_numbers = da.ones_like(ddid)
#    state_ids = da.zeros_like(ddid)
    
    chunksize_row = _calc_optimal_ms_chunk_shape(min_memory_for_a_thread,(n_row,),32,'ZERO')
    zero_column = da.zeros(n_row, chunks=chunksize_row, dtype="int32")

    chunksize_row = _calc_optimal_ms_chunk_shape(min_memory_for_a_thread,(n_row,),64,'EXPOSURE/INTERVAL')
    exposures = da.full((n_row,), time_xda.time_delta, chunks=chunksize_row, dtype="float64") # fill with input in units of the input array, which we
    '''


